import numpy as np
import pyarrow.compute as pc

from ...coordinates import CartesianCoordinates, CoordinateCovariances, Origin
from ...orbits import Orbits, VariantOrbits
from ...propagator import Propagator
from ...time import Timestamp
from ..impacts import (
    EarthImpacts,
    ImpactMixin,
    ImpactProbabilities,
    calculate_impact_probabilities,
    calculate_impacts,
    calculate_mahalanobis_distance,
)


class MockImpactPropagator(Propagator, ImpactMixin):
    def _propagate_orbits(self, orbits: Orbits, times: Timestamp) -> Orbits:
        return orbits

    def _detect_impacts(self, orbits: Orbits, num_days: float) -> Orbits:
        # Artificially set the orbits.coordinates.times to the end time
        # except for the orbits who impacted
        new_times = orbits.coordinates.time.add_days(num_days)
        orbits = orbits.set_column("coordinates.time", new_times)

        # Pick random orbit to impact
        impacted = orbits[0]
        impact = EarthImpacts.from_kwargs(
            orbit_id=impacted.orbit_id,
            variant_id=impacted.variant_id,
            coordinates=CartesianCoordinates.from_kwargs(
                x=impacted.coordinates.x,
                y=impacted.coordinates.y,
                z=impacted.coordinates.z,
                vx=impacted.coordinates.vx,
                vy=impacted.coordinates.vy,
                vz=impacted.coordinates.vz,
                time=impacted.coordinates.time,
                origin=impacted.coordinates.origin,
                frame=impacted.coordinates.frame,
            ),
            distance=[0.0],
        )

        return orbits, impact


def test_calculate_impacts():
    """
    Tests the i/o of calculate_impacts
    """
    orbits = Orbits.from_kwargs(
        orbit_id=["1", "2", "3"],
        object_id=["1", "2", "3"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
            covariance=CoordinateCovariances.from_sigmas(
                np.array(
                    [
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    ]
                )
            ),
        ),
    )

    prop = MockImpactPropagator()
    variants, impacts = calculate_impacts(
        orbits, 10, prop, num_samples=1000, processes=1, seed=42
    )
    assert pc.all(
        pc.equal(
            variants.coordinates.time.mjd().to_pylist(),
            orbits[0].coordinates.time.add_days(10).mjd().to_pylist()[0],
        )
    )


def test_calculate_impact_probabilities():
    """
    Tests the i/o of calculate_impact_probabilities
    """
    variants = VariantOrbits.from_kwargs(
        orbit_id=["1", "1", "1", "2", "2", "2", "3", "3", "3"],
        object_id=["1", "1", "1", "2", "2", "2", "3", "3", "3"],
        variant_id=["1", "2", "3", "1", "2", "3", "1", "2", "3"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0] * 3,
            y=[1.0, 2.0, 3.0] * 3,
            z=[1.0, 2.0, 3.0] * 3,
            vx=[1.0, 2.0, 3.0] * 3,
            vy=[1.0, 2.0, 3.0] * 3,
            vz=[1.0, 2.0, 3.0] * 3,
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 9),
            origin=Origin.from_kwargs(code=["SUN"] * 9),
            frame="ecliptic",
        ),
    )

    impacts = EarthImpacts.from_kwargs(
        orbit_id=["1", "2", "2"],
        variant_id=["1", "1", "2"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
        ),
        distance=[0.0, 0.0, 0.0],
    )

    ip = calculate_impact_probabilities(variants, impacts)

    assert ip == ImpactProbabilities.from_kwargs(
        orbit_id=["1", "2", "3"],
        impacts=[1, 2, 0],
        variants=[3, 3, 3],
        cumulative_probability=[1 / 3, 2 / 3, 0.0],
    )


def test_calculate_mahalanobis_distance():
    """ """
    observed_orbit = Orbits.from_kwargs(
        orbit_id=["1", "1", "1"],
        object_id=["1", "1", "1"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
            covariance=CoordinateCovariances.from_sigmas(
                np.array(
                    [
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    ]
                )
            ),
        ),
    )

    predicted_orbit = Orbits.from_kwargs(
        orbit_id=["1", "1", "1"],
        object_id=["1", "1", "1"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
        ),
    )

    md = calculate_mahalanobis_distance(observed_orbit, predicted_orbit)
    assert np.all(md == 0.0)

    less_perfect_observed_orbit = Orbits.from_kwargs(
        orbit_id=["1", "1", "1"],
        object_id=["1", "1", "1"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.1, 2.1, 3.1],
            y=[1.1, 2.1, 3.1],
            z=[1.1, 2.1, 3.1],
            vx=[1.1, 2.1, 3.1],
            vy=[1.1, 2.1, 3.1],
            vz=[1.1, 2.1, 3.1],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
            covariance=CoordinateCovariances.from_sigmas(
                np.array(
                    [
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    ]
                )
            ),
        ),
    )

    md = calculate_mahalanobis_distance(less_perfect_observed_orbit, predicted_orbit)
    np.testing.assert_almost_equal(md, np.array([0.24494897, 0.24494897, 0.24494897]))



import logging
import uuid
from typing import Iterable, Tuple

import numpy.typing as npt
import pyarrow.compute as pc
import quivr as qv

from ..coordinates.cartesian import CartesianCoordinates
from .classification import calc_orbit_class

logger = logging.getLogger(__name__)


class Orbits(qv.Table):

    orbit_id = qv.LargeStringColumn(default=lambda: uuid.uuid4().hex)
    object_id = qv.LargeStringColumn(nullable=True)
    coordinates = CartesianCoordinates.as_column()

    def group_by_orbit_id(self) -> Iterable[Tuple[str, "Orbits"]]:
        """
        Group orbits by orbit ID and yield them.

        Yields
        ------
        orbit_id : str
            Orbit ID.
        orbits : `~adam_core.orbits.orbits.Orbits`
            Orbits belonging to this orbit ID.
        """
        unique_orbit_ids = self.orbit_id.unique()
        for orbit_id in unique_orbit_ids:
            mask = pc.equal(self.orbit_id, orbit_id)
            yield orbit_id, self.apply_mask(mask)

    def dynamical_class(self) -> npt.NDArray[str]:
        """
        Compute dynamical classes of orbits. Currently
        limited to asteroid dynamical classes.

        Returns
        -------
        dynamical_classes : `~numpy.ndarray`
            Dynamical classes of orbits.
        """
        keplerian = self.coordinates.to_keplerian()
        return calc_orbit_class(keplerian)



from importlib.resources import files
from typing import Tuple

import numpy as np
import pandas as pd

from ...observations.associations import Associations
from ...observations.detections import PointSourceDetections
from ...observations.exposures import Exposures
from ...time import Timestamp


def make_observations() -> Tuple[Exposures, PointSourceDetections, Associations]:
    """
    Create an Exposures, PointSourceDetections, and Associations table
    using predicted ephemerides from JPL Horizons included in adam_core.

    Returns
    -------
    exposures : `~adam_core.observations.exposures.Exposures`
        Table of exposures.
    detections : `~adam_core.observations.detections.PointSourceDetections`
        Table of detections.
    associations : `~adam_core.observations.associations.Associations`
        Table of associations.
    """
    # Load the ephemeris (these are predicted ephemerides generated via JPL Horizons)
    ephemeris = pd.read_csv(
        files("adam_core.utils.helpers.data").joinpath("ephemeris.csv")
    )

    # Now lets create simulated exposures and detections
    exposure_times = {
        "X05": 30,
        "W84": 60,
    }
    for observatory_code in ephemeris["observatory_code"].unique():
        observatory_mask = ephemeris["observatory_code"] == observatory_code

        # For X05, lets give every observation a time thats between the start
        # and end time of an exposure. Note that with this current set up we
        # might get two exposures overlapping in time with different exposure start times
        if observatory_code == "X05":
            rng = np.random.default_rng(seed=20233202)
            observation_times = ephemeris[observatory_mask]["mjd_utc"].values
            exposure_start_times = observation_times + rng.uniform(
                -exposure_times[observatory_code] / 86400,
                0,
                size=len(observation_times),
            )
            ephemeris.loc[observatory_mask, "exposure_start"] = exposure_start_times

        # For everything else (which is just W84 at the moment) every observation is reported at the midpoint
        else:
            ephemeris.loc[observatory_mask, "exposure_start"] = ephemeris[
                observatory_mask
            ]["mjd_utc"].values - (exposure_times[observatory_code] / 2 / 86400)

    # Create an exposures table
    exposures = (
        ephemeris.groupby(by=["observatory_code", "exposure_start"])
        .size()
        .to_frame(name="num_obs")
        .reset_index()
    )
    exposures.sort_values(by="exposure_start", inplace=True, ignore_index=True)
    exposures.insert(
        0,
        "exposure_id",
        exposures["observatory_code"].astype(str)
        + "_"
        + [f"{i:04d}" for i in range(len(exposures))],
    )
    for observatory_code in exposures["observatory_code"].unique():
        exposures.loc[exposures["observatory_code"] == observatory_code, "duration"] = (
            exposure_times[observatory_code]
        )
    exposures["filter"] = "V"

    # Attached exposure IDs to the ephemerides
    ephemeris = ephemeris.merge(
        exposures[["observatory_code", "exposure_start", "exposure_id"]],
        on=["observatory_code", "exposure_start"],
    )

    # Create detections dataframe
    detections_df = ephemeris[
        ["exposure_id", "observatory_code", "mjd_utc", "RA", "DEC", "V", "targetname"]
    ].copy()
    detections_df.rename(
        columns={
            "RA": "ra",
            "DEC": "dec",
            "V": "mag",
        },
        inplace=True,
    )

    # Lets only report astrometric errors for one of the observatories
    detections_df.loc[detections_df["observatory_code"] == "X05", "ra_sigma"] = (
        0.1 / 3600
    )
    detections_df.loc[detections_df["observatory_code"] == "X05", "dec_sigma"] = (
        0.1 / 3600
    )
    # Lets report photometric errors for all observatories
    detections_df["mag_sigma"] = 0.1
    detections_df.sort_values(
        by=["mjd_utc", "observatory_code"], inplace=True, ignore_index=True
    )
    detections_df["detection_id"] = [f"obs_{i:04d}" for i in range(len(detections_df))]

    # Create exposures table
    exposures = Exposures.from_kwargs(
        id=exposures["exposure_id"],
        start_time=Timestamp.from_mjd(exposures["exposure_start"].values, scale="utc"),
        duration=exposures["duration"].values,
        filter=exposures["filter"].values,
        observatory_code=exposures["observatory_code"].values,
    )

    # Create detections table
    detections = PointSourceDetections.from_kwargs(
        id=detections_df["detection_id"],
        exposure_id=detections_df["exposure_id"],
        time=Timestamp.from_mjd(detections_df["mjd_utc"].values, scale="utc"),
        ra=detections_df["ra"].values,
        dec=detections_df["dec"].values,
        mag=detections_df["mag"].values,
        ra_sigma=detections_df["ra_sigma"].values,
        dec_sigma=detections_df["dec_sigma"].values,
        mag_sigma=detections_df["mag_sigma"].values,
    )

    # Create associations table
    associations = Associations.from_kwargs(
        detection_id=detections.id,
        object_id=detections_df["targetname"],
    )
    return exposures, detections, associations



from typing import Literal, Union

import numpy as np
import pyarrow.compute as pc
import spiceypy as sp

from ..constants import Constants as c
from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.origin import Origin, OriginCodes
from ..time import Timestamp
from ..utils.spice import get_perturber_state, setup_SPICE
from .observers import OBSERVATORY_CODES, OBSERVATORY_GEODETICS

R_EARTH = c.R_EARTH
OMEGA_EARTH = 2 * np.pi / 0.997269675925926
Z_AXIS = np.array([0, 0, 1])


def get_observer_state(
    code: Union[str, OriginCodes],
    times: Timestamp,
    frame: Literal["ecliptic", "equatorial"] = "ecliptic",
    origin: OriginCodes = OriginCodes.SUN,
) -> CartesianCoordinates:
    """
    Find state vectors for an observer at a given time in the given frame and measured from the given origin.

    For NAIF Origin Codes this function uses the SPICE kernels to find the state vectors.
    See `~adam_core.utils.spice.get_perturber_state` for more information.

    For MPC observatory codes this function currently only supports ground-based observers on Earth.
    In this case, the Earth body-fixed frame used for calculations is the standard ITRF93,
    which takes into account:

      - precession (IAU-1976)
      - nutation (IAU-1980 with IERS corrections)
      - polar motion

    This frame is retrieved through SPICE.

    Parameters
    ----------
    code : Union[str, OriginCodes]
        MPC observatory code or NAIF origin code for which to find the states.
    times : Timestamp (N)
        Epochs for which to find the observatory locations.
    frame : {'equatorial', 'ecliptic'}
        Return observer state in the equatorial or ecliptic J2000 frames.
    origin : OriginCodes
        The NAIF ID of the origin.

    Returns
    -------
    observer_states : `~adam_core.coordinates.cartesian.CartesianCoordinates`
        The state vectors of the observer in the desired frame
        and measured from the desired origin.
    """
    if isinstance(code, str) and code not in OBSERVATORY_CODES:
        err = f"{code} is not a valid MPC observatory code."
        raise ValueError(err)

    if frame == "ecliptic":
        frame_spice = "ECLIPJ2000"
    elif frame == "equatorial":
        frame_spice = "J2000"
    else:
        err = "frame should be one of {'equatorial', 'ecliptic'}"
        raise ValueError(err)

    # Make sure SPICE is ready to go
    setup_SPICE()

    # If the code is an OriginCode, we can just use the get_perturber_state function
    if isinstance(code, OriginCodes):
        return get_perturber_state(code, times, frame=frame, origin=origin)

    # If it not an origin code then lets get the state vector
    else:
        # Get observatory geodetic information
        geodetics = OBSERVATORY_GEODETICS.select("code", code)
        if len(geodetics) == 0:
            raise ValueError(
                f"Observatory code '{code}' not found in the observatory geodetics table."
            )
        geodetics = geodetics.table.to_pylist()[0]

        # Unpack geodetic information
        longitude = geodetics["longitude"]
        cos_phi = geodetics["cos_phi"]
        sin_phi = geodetics["sin_phi"]

        if np.any(np.isnan([longitude, cos_phi, sin_phi])):
            err = (
                f"{code} is missing information on Earth-based geodetic coordinates. The MPC Obs Code\n"
                "file may be missing this information or the observer is a space-based observatory.\n"
                "Space observatories are currently not supported.\n"
            )
            raise ValueError(err)

        # Grab Earth state vector (this function handles duplicate times)
        state = get_perturber_state(
            OriginCodes.EARTH, times, frame=frame, origin=origin
        )

        # If the code is 500 (geocenter), we can just return the Earth state vector
        if code == "500":
            return state

        # If not then we need to add a topocentric correction
        else:
            # Get observer location on Earth
            sin_longitude = np.sin(np.radians(longitude))
            cos_longitude = np.cos(np.radians(longitude))

            # Calculate pointing vector from geocenter to observatory
            o_hat_ITRF93 = np.array(
                [cos_longitude * cos_phi, sin_longitude * cos_phi, sin_phi]
            )

            # Multiply pointing vector with Earth radius to get actual vector
            o_vec_ITRF93 = np.dot(R_EARTH, o_hat_ITRF93)

            # Warning! Converting times to ET will incur a loss of precision.
            epochs_et = times.rescale("tdb").et()
            unique_epochs_et_tdb = epochs_et.unique()

            N = len(epochs_et)
            r_obs = np.empty((N, 3), dtype=np.float64)
            v_obs = np.empty((N, 3), dtype=np.float64)
            r_geo = state.r
            v_geo = state.v
            for epoch in unique_epochs_et_tdb:
                # Grab rotation matrices from ITRF93 to ecliptic J2000
                # The ITRF93 high accuracy Earth rotation model takes into account:
                # Precession:  1976 IAU model from Lieske.
                # Nutation:  1980 IAU model, with IERS corrections due to Herring et al.
                # True sidereal time using accurate values of TAI-UT1
                # Polar motion

                rotation_matrix = sp.pxform("ITRF93", frame_spice, epoch.as_py())

                # Find indices of epochs that match the current unique epoch
                mask = pc.equal(epochs_et, epoch).to_numpy(False)

                # Add o_vec + r_geo to get r_obs (thank you numpy broadcasting)
                r_obs[mask] = r_geo[mask] + rotation_matrix @ o_vec_ITRF93

                # Calculate the velocity (thank you numpy broadcasting)
                rotation_direction = np.cross(o_hat_ITRF93, Z_AXIS)
                v_obs[mask] = v_geo[mask] + rotation_matrix @ (
                    -OMEGA_EARTH * R_EARTH * rotation_direction
                )

        return CartesianCoordinates.from_kwargs(
            time=times,
            x=r_obs[:, 0],
            y=r_obs[:, 1],
            z=r_obs[:, 2],
            vx=v_obs[:, 0],
            vy=v_obs[:, 1],
            vz=v_obs[:, 2],
            frame=frame,
            origin=Origin.from_kwargs(code=[origin.name for i in range(len(times))]),
        )



from astroquery.jplhorizons import Horizons

from adam_core.coordinates.cartesian import CartesianCoordinates
from adam_core.coordinates.covariances import CoordinateCovariances
from adam_core.coordinates.origin import Origin, OriginCodes
from adam_core.time import Timestamp

observatory_codes = ["I41", "X05", "F51", "W84", "000", "500"]

for code in observatory_codes:
    for id in ["sun", "ssb"]:
        if id == "sun":
            origin = OriginCodes.SUN
        else:
            origin = OriginCodes.SOLAR_SYSTEM_BARYCENTER

        horizons = Horizons(
            id=id,
            location=code,
            epochs={"start": "2013-01-01", "stop": "2023-01-01", "step": "10d"},
        )
        result = horizons.vectors(
            refplane="ecliptic", aberrations="geometric"
        ).to_pandas()

        # Flip the signs of the state to get the state of the observer
        states = CartesianCoordinates.from_kwargs(
            time=Timestamp.from_jd(result["datetime_jd"].values, scale="tdb"),
            covariance=CoordinateCovariances.nulls(len(result)),
            x=-result["x"].values,
            y=-result["y"].values,
            z=-result["z"].values,
            vx=-result["vx"].values,
            vy=-result["vy"].values,
            vz=-result["vz"].values,
            origin=Origin.from_kwargs(code=[origin.name for _ in range(len(result))]),
            frame="ecliptic",
        )
        states.to_parquet(f"{code}_{id}.parquet")



from typing import Tuple

import jax.numpy as jnp
import jax.typing as jnpt
from jax import config, jit, lax

from .barker import solve_barker

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")


@jit
def calc_period(a: jnpt.ArrayLike, mu: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """
    Calculate the period of an orbit given the semi-major axis and
    gravitational parameter.

    Parameters
    ----------
    a
        Semi-major axis.
    mu
        Gravitational parameter.

    Returns
    -------
    P
        Period.
    """
    return jnp.where(a < 0.0, jnp.inf, 2 * jnp.pi * jnp.sqrt(a**3 / mu))


@jit
def calc_periapsis_distance(a: jnpt.ArrayLike, e: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """
    Calculate the periapsis distance of an orbit given the semi-major axis and
    eccentricity.

    Parameters
    ----------
    a
        Semi-major axis.
    e
        Eccentricity.

    Returns
    -------
    q
        Periapsis distance.
    """
    return a * (1 - e)


@jit
def calc_apoapsis_distance(a: jnpt.ArrayLike, e: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """
    Calculate the apoapsis distance of an orbit given the semi-major axis and
    eccentricity.

    Parameters
    ----------
    a
        Semi-major axis.
    e
        Eccentricity.

    Returns
    -------
    Q
        Apoapsis distance.
    """
    return jnp.where(e >= 1.0, jnp.inf, a * (1 + e))


@jit
def calc_semi_major_axis(q: jnpt.ArrayLike, e: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """
    Calculate the semi-major axis of an orbit given the periapsis distance and
    eccentricity.

    Parameters
    ----------
    q
        Periapsis distance.
    e
        Eccentricity.

    Returns
    -------
    a
        Semi-major axis.
    """
    return q / (1 - e)


@jit
def calc_semi_latus_rectum(a: jnpt.ArrayLike, e: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """
    Calculate the semi-latus rectum of an orbit given the semi-major axis and
    eccentricity.

    Parameters
    ----------
    a
        Semi-major axis.
    e
        Eccentricity.

    Returns
    -------
    p
        Semi-latus rectum.
    """
    return a * (1 - e**2)


@jit
def calc_mean_motion(a: jnpt.ArrayLike, mu: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """
    Calculate the mean motion of an orbit given the semi-major axis and
    gravitational parameter.

    Parameters
    ----------
    a
        Semi-major axis.
    mu
        Gravitational parameter.

    Returns
    -------
    n
        Mean motion in radians per unit time.
    """
    return jnp.sqrt(mu / jnp.abs(a) ** 3)


@jit
def calc_mean_anomaly(nu: float, e: float) -> float:
    """
    Calculate the mean anomaly given true anomaly in radians
    and eccentricity.

    Parameters
    ----------
    nu : float
        True anomaly in radians.
    e : float
        Eccentricity.

    Returns
    -------
    M : float
        Mean anomaly in radians.
    """
    E, M = lax.cond(
        e < 1.0,
        _calc_elliptical_anomalies,
        lambda nu, e: lax.cond(
            e > 1.0, _calc_hyperbolic_anomalies, _calc_parabolic_anomalies, nu, e
        ),
        nu,
        e,
    )

    return M


@jit
def _calc_elliptical_anomalies(nu: float, e: float) -> Tuple[float, float]:
    nu_ = jnp.where(nu >= 2 * jnp.pi, nu % (2 * jnp.pi), nu)
    E = jnp.arctan2(jnp.sqrt(1 - e**2) * jnp.sin(nu_), e + jnp.cos(nu_))
    M = E - e * jnp.sin(E)
    M = jnp.where(M < 0.0, M + 2 * jnp.pi, M)
    return E, M


@jit
def _calc_hyperbolic_anomalies(nu: float, e: float) -> Tuple[float, float]:
    nu_ = jnp.where(nu >= 2 * jnp.pi, nu % (2 * jnp.pi), nu)
    H = 2 * jnp.arctanh(jnp.sqrt((e - 1) / (e + 1)) * jnp.tan(nu_ / 2))
    M = e * jnp.sinh(H) - H
    M = jnp.where(M < 0.0, M + 2 * jnp.pi, M)
    return H, M


@jit
def _calc_parabolic_anomalies(nu: float, e: float) -> Tuple[float, float]:
    nu_ = jnp.where(nu >= 2 * jnp.pi, nu % (2 * jnp.pi), nu)
    D = jnp.arctan(nu_ / 2)
    M = D + (D**3 / 3)
    M = jnp.where(M < 0.0, M + 2 * jnp.pi, M)
    return D, M


@jit
def solve_kepler(e: float, M: float, max_iter: int = 100, tol: float = 1e-15) -> float:
    """
    Solve Kepler's equation for true anomaly (nu) given eccentricity
    and mean anomaly using Newton-Raphson. Technically, this is only valid for orbits
    with eccentricity < 1.0 and eccentricity > 1.0. However, for parabolic orbits (e = 1.0)
    this function will call the `solve_barker` function from `thor.dynamics.barker`.

    Parameters
    ----------
    e : float
        Eccentricity
    M : float
        Mean anomaly in radians.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will use the value of the relevant anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute anomalies using the Newtown-Raphson
        method.

    Returns
    -------
    nu : float
        True anomaly in radians.

    References
    ----------
    [1] Curtis, H. D. (2014). Orbital Mechanics for Engineering Students. 3rd ed.,
        Elsevier Ltd. ISBN-13: 978-0080977478
    """
    ratio = 1e15
    iterations = 0
    E_init = jnp.where(e < 1.0, M, M)

    p = [E_init, e, M, ratio, iterations]

    @jit
    def _elliptical_newton_raphson(p):
        E = p[0]
        e = p[1]
        M = p[2]
        iterations = p[4]

        # Newton-Raphson
        # Equation 3.17 in Curtis (2014) [1]
        f = E - e * jnp.sin(E) - M
        fp = 1 - e * jnp.cos(E)

        ratio = f / fp
        E -= ratio
        iterations += 1

        p[0] = E
        p[1] = e
        p[2] = M
        p[3] = ratio
        p[4] = iterations
        return p

    @jit
    def _hyperbolic_newton_raphson(p):
        F = p[0]
        e = p[1]
        M = p[2]
        iterations = p[4]

        # Newton-Raphson
        # Equation 3.45 in Curtis (2014) [1]
        f = e * jnp.sinh(F) - F - M
        fp = e * jnp.cosh(F) - 1

        ratio = f / fp
        F -= ratio
        iterations += 1

        p[0] = F
        p[1] = e
        p[2] = M
        p[3] = ratio
        p[4] = iterations
        return p

    # Define while loop condition function
    @jit
    def _while_condition(p):
        ratio = p[-2]
        iterations = p[-1]
        return (jnp.abs(ratio) > tol) & (iterations <= max_iter)

    # Calculate parameters, if e < 1.0 then the orbit is elliptical
    # if e > 1.0 then the orbit is hyperbolic
    p = lax.cond(
        e < 1.0,
        lambda p: lax.while_loop(_while_condition, _elliptical_newton_raphson, p),
        lambda p: lax.cond(
            e > 1.0,
            lambda p: lax.while_loop(
                _while_condition,
                _hyperbolic_newton_raphson,
                p,
            ),
            # For parabolic orbits return the parameters as is since
            # no iteration is needed for parabolic orbits
            lambda p: p,
            p,
        ),
        p,
    )

    nu = lax.cond(
        e < 1.0,
        lambda E, e, M: 2
        * jnp.arctan2(
            jnp.sqrt(1 + e) * jnp.sin(E / 2), jnp.sqrt(1 - e) * jnp.cos(E / 2)
        ),
        lambda E, e, M: lax.cond(
            e > 1.0,
            lambda H, e, M: 2
            * jnp.arctan(
                jnp.sqrt(e + 1) * jnp.sinh(H / 2) / (jnp.sqrt(e - 1) * jnp.cosh(H / 2))
            ),
            lambda P, e, M: solve_barker(M),
            p[0],
            p[1],
            p[2],
        ),
        p[0],
        p[1],
        p[2],
    )

    # True anomaly should be in the range [0, 2*pi)
    nu = jnp.where(nu < 0.0, nu + 2 * jnp.pi, nu)
    nu = jnp.where(nu >= 2 * jnp.pi, nu % (2 * jnp.pi), nu)
    return nu



from adam_core.ray_cluster import initialize_use_ray


# Test that we are calling the correct ray init function using mocks
def test_initialize_ray_no_cluster(mocker):
    mock_ray = mocker.patch("adam_core.ray_cluster.ray")

    # Ensure we initialize ray with the correct values
    mock_ray.is_initialized.return_value = False
    # Set mock to throw an error if init is called with address="auto"
    # but not if called with other args
    mock_ray.init.side_effect = [
        ConnectionError,
        None,
    ]

    initialize_use_ray(num_cpus=4, object_store_bytes=1000)
    mock_ray.init.assert_called_with(
        dashboard_host="0.0.0.0", num_cpus=4, object_store_memory=1000
    )
    mock_ray.init.call_count == 1


def test_initialize_ray_existing_cluster(mocker):
    mock_ray = mocker.patch("adam_core.ray_cluster.ray")

    mock_ray.is_initialized.return_value = False
    mock_ray.init.return_value = True
    initialize_use_ray()

    mock_ray.init.assert_called_once_with(address="auto")
    assert mock_ray.init.call_count == 1


def test_initialize_ray_fully_initialized(mocker):
    mock_ray = mocker.patch("adam_core.ray_cluster.ray")
    mock_ray.is_initialized.return_value = True
    initialize_use_ray()
    assert mock_ray.init.call_count == 0



from typing import List, Optional, Tuple, Union

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv

from ..coordinates.residuals import Residuals, calculate_reduced_chi2
from ..coordinates.spherical import SphericalCoordinates
from ..observers.observers import Observers
from ..orbits.orbits import Orbits
from ..propagator.propagator import Propagator
from .fitted_orbits import FittedOrbitMembers, FittedOrbits


class OrbitDeterminationObservations(qv.Table):

    id = qv.LargeStringColumn()
    coordinates = SphericalCoordinates.as_column()
    observers = Observers.as_column()


def evaluate_orbits(
    orbits: Union[Orbits, FittedOrbits],
    observations: OrbitDeterminationObservations,
    propagator: Propagator,
    parameters: int = 6,
    ignore: Optional[List[str]] = None,
) -> Tuple["FittedOrbits", "FittedOrbitMembers"]:
    """
    Creates a fitted orbit and fitted orbit members from input orbits and observations.
    This function takes the input orbits and calculates the residuals with respect to the observations.
    It then computes the chi2 and reduced chi2 values for the orbits. If outliers are provided, they are ignored when calculating
    the chi2, reduced chi2, and arc length values.

    This function is intended to be used to evaluate the quality of orbits with respect to observations
    in scenarios for orbit determination.

    Parameters
    ----------
    orbit : `~adam_core.orbits.Orbits` (N)
        Orbits to calculate residuals with respect to the observations for.
    observations : `~adam_core.orbit_determination.DifferentialCorrectionObservations` (M)
        Observations believed to belong to the input orbit.
    propagator : `~adam_core.propagator.Propagator`
        Propagator to use to generate ephemeris.
    parameters : int
        Number of parameters that were initially fit to the observations. This is typically
        6 for an orbit fit to observations (assuming the epoch was not fit).
    ignore : list of str
        List of observation IDs to ignore when calculating chi2 and reduced chi2 values. This
        is typically a list of outlier observation IDs.

    Returns
    -------
    fitted_orbit : `~adam_core.orbit_determination.FittedOrbits` (N)
        Fitted orbit.
    fitted_orbit_members : `~adam_core.orbit_determination.FittedOrbitMembers` (N * M)
        Fitted orbit members.
    """
    num_orbits = len(orbits)
    if isinstance(orbits, FittedOrbits):
        orbits = orbits.to_orbits()

    assert len(orbits) == len(orbits.orbit_id.unique())

    # Compute ephemeris and residuals
    ephemeris = propagator.generate_ephemeris(
        orbits,
        observations.observers,
        max_processes=1,
    )

    # Sort the orbits by ID (the ephemeris is already sorted by
    # orbit ID, time, and origin)
    orbits = orbits.sort_by(["orbit_id"])

    # Stack the observations into a single table and compute the residuals with respect
    # to the predicted coordinates
    observations_stacked = qv.concatenate([observations for _ in range(num_orbits)])
    residuals = Residuals.calculate(
        observations_stacked.coordinates, ephemeris.coordinates
    )

    # If outliers are provided, we need to mask them out of the observations
    # before we compute the chi2, reduced chi2, and arc length values.
    if ignore is not None:
        mask = pc.invert(pc.is_in(observations.id, pa.array(ignore)))
        observations_to_include = observations.apply_mask(mask)
    else:
        mask = pa.repeat(True, len(observations))
        observations_to_include = observations

    # Compute number of observations
    num_obs = len(observations_to_include)

    # Compute chi2 and reduced chi2 for each orbit
    chi2 = np.empty(num_orbits, dtype=np.float64)
    reduced_chi2 = np.empty(num_orbits, dtype=np.float64)
    for i, orbit_id in enumerate(orbits.orbit_id):
        orbit_mask = pc.equal(ephemeris.orbit_id, orbit_id)
        residuals_to_include = residuals.apply_mask(orbit_mask).apply_mask(mask)
        chi2[i] = pc.sum(residuals_to_include.chi2).as_py()
        reduced_chi2[i] = calculate_reduced_chi2(residuals_to_include, parameters)

    # Compute arc length for the orbits (will be the same for all orbits)
    arc_length = (
        observations_to_include.coordinates.time.max().mjd()[0].as_py()
        - observations_to_include.coordinates.time.min().mjd()[0].as_py()
    )

    # Now we create a fitted orbit and fitted orbit members from the solution orbit
    # and residuals. We also need to compute the chi2 and reduced chi2 values.
    fitted_orbit = FittedOrbits.from_kwargs(
        orbit_id=orbits.orbit_id,
        object_id=orbits.object_id,
        coordinates=orbits.coordinates,
        arc_length=pa.repeat(arc_length, num_orbits),
        num_obs=pa.repeat(num_obs, num_orbits),
        chi2=chi2,
        reduced_chi2=reduced_chi2,
    )
    fitted_orbit_members = FittedOrbitMembers.from_kwargs(
        orbit_id=ephemeris.orbit_id,
        obs_id=observations_stacked.id,
        residuals=residuals,
        outlier=pa.concat_arrays([pc.invert(mask) for i in range(num_orbits)]),
    )

    return fitted_orbit, fitted_orbit_members



from importlib.resources import files

import pandas as pd
import pytest


@pytest.fixture
def orbital_elements():
    orbital_elements_file = files("adam_core.utils.helpers.data").joinpath(
        "elements_sun_ec.csv"
    )
    df = pd.read_csv(
        orbital_elements_file, index_col=False, float_precision="round_trip"
    )
    return df


@pytest.fixture
def orbital_elements_equatorial():
    orbital_elements_file = files("adam_core.utils.helpers.data").joinpath(
        "elements_sun_eq.csv"
    )
    df = pd.read_csv(
        orbital_elements_file, index_col=False, float_precision="round_trip"
    )
    return df


@pytest.fixture
def orbital_elements_barycentric():
    orbital_elements_file = files("adam_core.utils.helpers.data").joinpath(
        "elements_ssb_ec.csv"
    )
    df = pd.read_csv(
        orbital_elements_file, index_col=False, float_precision="round_trip"
    )
    return df


@pytest.fixture
def orbital_elements_barycentric_equatorial():
    orbital_elements_file = files("adam_core.utils.helpers.data").joinpath(
        "elements_ssb_eq.csv"
    )
    df = pd.read_csv(
        orbital_elements_file, index_col=False, float_precision="round_trip"
    )
    return df



from typing import Tuple

import jax.numpy as jnp
from jax import config, jit, lax, vmap

from ..constants import Constants as c
from .propagation import _propagate_2body

config.update("jax_enable_x64", True)

MU = c.MU
C = c.C


@jit
def _add_light_time(
    orbit: jnp.ndarray,
    t0: float,
    observer_position: jnp.ndarray,
    lt_tol: float = 1e-10,
    mu: float = MU,
    max_iter: int = 1000,
    tol: float = 1e-15,
) -> Tuple[jnp.ndarray, jnp.float64]:
    """
    When generating ephemeris, orbits need to be backwards propagated to the time
    at which the light emitted or relflected from the object towards the observer.

    Light time correction must be added to orbits in expressed in an inertial frame (ie, orbits
    must be barycentric).

    Parameters
    ----------
    orbit : `~jax.numpy.ndarray` (6)
        Barycentric orbit in cartesian elements to correct for light time delay.
    t0 : float
        Epoch at which orbits are defined.
    observer_positions : `~jax.numpy.ndarray` (3)
        Location of the observer in barycentric cartesian elements at the time of observation.
    lt_tol : float, optional
        Calculate aberration to within this value in time (units of days.)
    mu : float, optional
        Gravitational parameter (GM) of the attracting body in units of
        AU**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge for propagation.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly during propagation using the Newtown-Raphson
        method.

    Returns
    -------
    corrected_orbit : `~jax.numpy.ndarray` (6)
        Orbit adjusted for light travel time.
    lt : float
        Light time correction (t0 - corrected_t0).
    """
    dlt = 1e30
    lt = 1e30

    @jit
    def _iterate_light_time(p):

        orbit_i = p[0]
        t0 = p[1]
        lt0 = p[2]
        dlt = p[3]

        # Calculate topocentric distance
        rho = jnp.linalg.norm(orbit_i[:3] - observer_position)

        # Calculate initial guess of light time
        lt = rho / C

        # Calculate difference between previous light time correction
        # and current guess
        dlt = jnp.abs(lt - lt0)

        # Propagate backwards to new epoch
        t1 = t0 - lt
        orbit_propagated = _propagate_2body(
            orbit, t0, t1, mu=mu, max_iter=max_iter, tol=tol
        )

        p[0] = orbit_propagated
        p[1] = t1
        p[2] = lt
        p[3] = dlt
        return p

    @jit
    def _while_condition(p):
        dlt = p[-1]
        return dlt > lt_tol

    p = [orbit, t0, lt, dlt]
    p = lax.while_loop(_while_condition, _iterate_light_time, p)

    orbit_aberrated = p[0]
    t0_aberrated = p[1]  # noqa: F841
    lt = p[2]
    return orbit_aberrated, lt


# Vectorization Map: _add_light_time
_add_light_time_vmap = jit(
    vmap(_add_light_time, in_axes=(0, 0, 0, None, None, None, None), out_axes=(0, 0))
)


@jit
def add_light_time(
    orbits: jnp.ndarray,
    t0: jnp.ndarray,
    observer_positions: jnp.ndarray,
    lt_tol: float = 1e-10,
    mu: float = MU,
    max_iter: int = 1000,
    tol: float = 1e-15,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    When generating ephemeris, orbits need to be backwards propagated to the time
    at which the light emitted or relflected from the object towards the observer.

    Light time correction must be added to orbits in expressed in an inertial frame (ie, orbits
    must be barycentric).

    Parameters
    ----------
    orbits : `~jax.numpy.ndarray` (N, 6)
        Barycentric orbits in cartesian elements to correct for light time delay.
    t0 : `~jax.numpy.ndarray` (N)
        Epoch at which orbits are defined.
    observer_positions : `~jax.numpy.ndarray` (N, 3)
        Location of the observer in barycentric cartesian elements at the time of observation.
    lt_tol : float, optional
        Calculate aberration to within this value in time (units of days.)
    mu : float, optional
        Gravitational parameter (GM) of the attracting body in units of
        AU**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge for propagation.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly during propagation using the Newtown-Raphson
        method.

    Returns
    -------
    corrected_orbits : `~jax.numpy.ndarray` (N, 6)
        Orbits adjusted for light travel time.
    lt : `~jax.numpy.ndarray` (N)
        Light time correction (t0 - corrected_t0).
    """
    orbits_aberrated, lts = _add_light_time_vmap(
        orbits, t0, observer_positions, lt_tol, mu, max_iter, tol
    )
    return orbits_aberrated, lts


@jit
def add_stellar_aberration(
    orbits: jnp.ndarray, observer_states: jnp.ndarray
) -> jnp.ndarray:
    """
    The motion of the observer in an inertial frame will cause an object
    to appear in a different location than its true geometric location. This
    aberration is typically applied after light time corrections have been added.

    The velocity of the input orbits are unmodified only the position
    vector is modified with stellar aberration.

    Parameters
    ----------
    orbits : `~jax.numpy.ndarray` (N, 6)
        Orbits in barycentric cartesian elements.
    observer_states : `~jax.numpy.ndarray` (N, 6)
        Observer states in barycentric cartesian elements.

    Returns
    -------
    rho_aberrated : `~jax.numpy.ndarray` (N, 3)
        The topocentric position vector for each orbit with
        added stellar aberration.

    References
    ----------
    [1] Urban, S. E; Seidelmann, P. K. (2013) Explanatory Supplement to the Astronomical Almanac. 3rd ed.,
        University Science Books. ISBN-13: 978-1891389856
    """
    topo_states = orbits - observer_states
    rho_aberrated = jnp.zeros((len(topo_states), 3), dtype=jnp.float64)
    rho_aberrated = rho_aberrated.at[:].set(topo_states[:, :3])

    v_obs = observer_states[:, 3:]
    gamma = v_obs / C
    beta_inv = jnp.sqrt(1 - jnp.linalg.norm(gamma, axis=1, keepdims=True) ** 2)
    delta = jnp.linalg.norm(topo_states[:, :3], axis=1, keepdims=True)

    # Equation 7.40 in Urban & Seidelmann (2013) [1]
    rho = topo_states[:, :3] / delta
    rho_dot_gamma = jnp.sum(rho * gamma, axis=1, keepdims=True)
    rho_aberrated = rho_aberrated.at[:].set(
        (beta_inv * rho + gamma + rho_dot_gamma * gamma / (1 + beta_inv))
        / (1 + rho_dot_gamma)
    )
    rho_aberrated = rho_aberrated.at[:].multiply(delta)

    return rho_aberrated



import numpy as np
import spiceypy as sp
from naif_leapseconds import leapseconds

from ...time import Timestamp
from ..spice import _jd_tdb_to_et


def test__jd_tdb_to_et():
    # Test that _jd_tdb_to_et returns the same values as SPICE's str2et
    sp.furnsh(leapseconds)

    times = Timestamp.from_mjd(np.arange(40000, 70000, 5), scale="tdb")
    jd_tdb = times.jd().to_numpy()

    et_actual = _jd_tdb_to_et(jd_tdb)
    et_expected = np.array([sp.str2et(f"JD {i:.16f} TDB") for i in jd_tdb])

    np.testing.assert_equal(et_actual, et_expected)



# flake8: noqa: F401
from .differential_correction import fit_least_squares
from .evaluate import OrbitDeterminationObservations, evaluate_orbits
from .fitted_orbits import FittedOrbitMembers, FittedOrbits
from .outliers import remove_lowest_probability_observation



from importlib.resources import files

import pandas as pd
import pytest

from adam_core.orbits import Orbits


@pytest.fixture
def orbital_elements():
    orbital_elements_file = files("adam_core.utils.helpers.data").joinpath(
        "elements_sun_ec.csv"
    )
    df = pd.read_csv(
        orbital_elements_file, index_col=False, float_precision="round_trip"
    )
    return df


@pytest.fixture
def orbital_elements_barycentric():
    orbital_elements_file = files("adam_core.utils.helpers.data").joinpath(
        "elements_ssb_ec.csv"
    )
    df = pd.read_csv(
        orbital_elements_file, index_col=False, float_precision="round_trip"
    )
    return df


@pytest.fixture
def propagated_orbits():
    return Orbits.from_parquet(
        files("adam_core.utils.helpers.data").joinpath("propagated_orbits.parquet")
    )


@pytest.fixture
def ephemeris():
    ephemeris_file = files("adam_core.utils.helpers.data").joinpath("ephemeris.csv")
    df = pd.read_csv(
        ephemeris_file,
        index_col=False,
        float_precision="round_trip",
        dtype={"orbit_id": str},
    )
    return df



import healpy

from ...time import Timestamp
from ..detections import PointSourceDetections
from ..exposures import Exposures


def test_detections_link_to_exposures():
    start_times = Timestamp.from_iso8601(
        [
            "2000-01-01T00:00:00",
            "2000-01-02T00:00:00",
        ],
        scale="utc",
    )
    exp = Exposures.from_kwargs(
        id=["e1", "e2"],
        start_time=start_times,
        duration=[60, 30],
        filter=["g", "r"],
        observatory_code=["I41", "I41"],
    )

    det = PointSourceDetections.from_kwargs(
        id=["d1", "d2", "d3", "d4", "d5"],
        exposure_id=["e1", "e1", "e2", "e1", "e2"],
        time=None,
        ra=[0, 1, 2, 3, 4],
        dec=[5, 6, 7, 8, 9],
        mag=[10, 11, 12, 13, 14],
    )

    link = det.link_to_exposures(exp)

    # one link per unique exposure id
    assert len(link) == 2

    have_det, have_exp = link.select("e1")
    assert len(have_exp) == 1
    assert len(have_det) == 3
    assert have_det.id.to_pylist() == ["d1", "d2", "d4"]

    have_det, have_exp = link.select("e2")
    assert len(have_exp) == 1
    assert len(have_det) == 2
    assert have_det.id.to_pylist() == ["d3", "d5"]


def test_detection_healpixels():
    ra, dec = healpy.pixelfunc.pix2ang(
        nside=16, ipix=[1, 2, 3, 4, 5], nest=True, lonlat=True
    )
    det = PointSourceDetections.from_kwargs(
        id=["d1", "d2", "d3", "d4", "d5"],
        exposure_id=["e1", "e1", "e2", "e1", "e2"],
        time=None,
        ra=ra,
        dec=dec,
        mag=[10, 11, 12, 13, 14],
    )

    healpixels = det.healpixels(nside=16, nest=True)

    assert len(healpixels) == 5
    assert healpixels.tolist() == [1, 2, 3, 4, 5]


def test_detection_group_by_healpixel():
    healpixels = [3, 1, 1, 2, 1, 3, 1, 2]
    ids = ["d1", "d2", "d3", "d4", "d5", "d6", "d7", "d8"]
    ra, dec = healpy.pixelfunc.pix2ang(
        nside=16, ipix=healpixels, nest=True, lonlat=True
    )
    det = PointSourceDetections.from_kwargs(
        id=ids,
        exposure_id=[""] * 8,
        ra=ra,
        dec=dec,
        mag=[0] * 8,
    )

    groups = dict(det.group_by_healpixel(nside=16, nest=True))
    assert len(groups) == 3

    assert 1 in groups
    assert len(groups[1]) == 4
    assert groups[1].id.to_pylist() == ["d2", "d3", "d5", "d7"]

    assert 2 in groups
    assert len(groups[2]) == 2
    assert groups[2].id.to_pylist() == ["d4", "d8"]

    assert 3 in groups
    assert len(groups[3]) == 2
    assert groups[3].id.to_pylist() == ["d1", "d6"]


def test_detections_group_by_exposure():
    detections = PointSourceDetections.from_kwargs(
        id=["d1", "d2", "d3", "d4", "d5"],
        exposure_id=["e1", "e1", "e2", "e1", "e2"],
        time=None,
        ra=[0, 1, 2, 3, 4],
        dec=[5, 6, 7, 8, 9],
        mag=[10, 11, 12, 13, 14],
    )

    groups = list(detections.group_by_exposure())
    assert len(groups) == 2

    assert groups[0].id.to_pylist() == ["d1", "d2", "d4"]
    assert groups[0].exposure_id.to_pylist() == ["e1", "e1", "e1"]
    assert groups[0].ra.to_pylist() == [0, 1, 3]
    assert groups[0].dec.to_pylist() == [5, 6, 8]

    assert groups[1].id.to_pylist() == ["d3", "d5"]
    assert groups[1].exposure_id.to_pylist() == ["e2", "e2"]
    assert groups[1].ra.to_pylist() == [2, 4]
    assert groups[1].dec.to_pylist() == [7, 9]



from __future__ import annotations

from typing import Generic, Literal, Optional, Protocol, TypeVar

import numpy as np
import pyarrow as pa
import quivr as qv

from . import types
from .cartesian import CartesianCoordinates
from .cometary import CometaryCoordinates
from .covariances import (
    sample_covariance_random,
    sample_covariance_sigma_points,
    weighted_covariance,
    weighted_mean,
)
from .keplerian import KeplerianCoordinates
from .spherical import SphericalCoordinates

T = TypeVar("T", bound=types.CoordinateType, covariant=True)


class VariantCoordinatesTable(Generic[T], Protocol):
    """
    A protocol for a generic table of variant coordinates.
    """

    @property
    def index(self) -> pa.Int64Array: ...

    @property
    def sample(self) -> T: ...

    @property
    def weight(self) -> pa.lib.DoubleArray: ...

    @property
    def weight_cov(self) -> pa.lib.DoubleArray: ...


def create_coordinate_variants(
    coordinates: types.CoordinateType,
    method: Literal["auto", "sigma-point", "monte-carlo"] = "auto",
    num_samples: int = 10000,
    alpha: float = 1,
    beta: float = 0,
    kappa: float = 0,
    seed: Optional[int] = None,
) -> VariantCoordinatesTable[types.CoordinateType]:
    """
    Sample and create variants for the given coordinates by sampling the covariance matrices.
    There are three supported methods:

    1. sigma-point: Sample the covariance matrix using sigma points. This is the fastest method,
       but can be inaccurate if the covariance matrix is not well behaved.
    2. monte-carlo: Sample the covariance matrix using a monte carlo method.
       This is the slowest method, but is the most accurate.
    3. auto: Automatically select the best method based on the covariance matrix.
       If the covariance matrix is well behaved then sigma-point sampling will be used.
       If the covariance matrix is not well behaved then monte-carlo sampling will be used.

    When sampling with monte-carlo, 10k samples are drawn. Sigma-point sampling draws 13 samples
    for 6-dimensional coordinates.

    .. warning::

        This function does not yet handle sampling of covariances and coordinates with missing values.

    Parameters
    ----------
    coordinates :
        The coordinates to sample.
    method:
        The method to use for sampling the covariance matrix. If 'auto' is selected then the method
        will be automatically selected based on the covariance matrix. The default is 'auto'.
    num_samples : int, optional
        The number of samples to draw when sampling with monte-carlo.
    alpha:
        Spread of the sigma points between 1e^-2 and 1.
    beta:
        Prior knowledge of the distribution when generating sigma points usually set to 2 for a Gaussian.
    kappa:
        Secondary scaling parameter when generating sigma points usually set to 0.


    Returns
    -------
        The variant coordinates.

    Raises
    ------
    ValueError:
        If the covariance matrices are all undefined.
        If the input coordinates are not supported.
    """

    if coordinates.covariance.is_all_nan():
        raise ValueError(
            "Cannot sample coordinate covariances when covariances are all undefined."
        )

    class VariantCoordinates(qv.Table):
        index = qv.Int64Column()
        sample = coordinates.as_column()
        weight = qv.Float64Column()
        weight_cov = qv.Float64Column()

    if isinstance(coordinates, CartesianCoordinates):
        dimensions = ["x", "y", "z", "vx", "vy", "vz"]
    elif isinstance(coordinates, SphericalCoordinates):
        dimensions = ["rho", "lon", "lat", "vrho", "vlon", "vlat"]
    elif isinstance(coordinates, KeplerianCoordinates):
        dimensions = ["a", "e", "i", "raan", "ap", "M"]
    elif isinstance(coordinates, CometaryCoordinates):
        dimensions = ["q", "e", "i", "raan", "ap", "tp"]
    else:
        raise ValueError(f"Unsupported coordinate type: {type(coordinates)}")

    variants_list = []
    for i, coordinate_i in enumerate(coordinates):

        mean = coordinate_i.values[0]
        cov = coordinate_i.covariance.to_matrix()[0]

        if np.any(np.isnan(cov)):
            raise ValueError(
                "Cannot sample coordinate covariances when some covariance elements are undefined."
            )
        if np.any(np.isnan(mean)):
            raise ValueError(
                "Cannot sample coordinate covariances when some coordinate dimensions are undefined."
            )

        if method == "sigma-point":
            samples, W, W_cov = sample_covariance_sigma_points(
                mean, cov, alpha=alpha, beta=beta, kappa=kappa
            )

        elif method == "monte-carlo":
            samples, W, W_cov = sample_covariance_random(
                mean, cov, num_samples=num_samples, seed=seed
            )

        elif method == "auto":
            # Sample with sigma points
            samples, W, W_cov = sample_covariance_sigma_points(
                mean, cov, alpha=alpha, beta=beta, kappa=kappa
            )

            # Check if the sigma point sampling is good enough by seeing if we can
            # recover the mean and covariance from the sigma points
            mean_sg = weighted_mean(samples, W)
            cov_sg = weighted_covariance(mean_sg, samples, W_cov)

            # If the sigma point sampling is not good enough, then sample with monte carlo
            # Though it is not guaranteed that monte carlo will actually be better
            diff_mean = np.abs(mean_sg - mean)
            diff_cov = np.abs(cov_sg - cov)
            if np.any(diff_mean >= 1e-12) or np.any(diff_cov >= 1e-12):
                samples, W, W_cov = sample_covariance_random(
                    mean, cov, num_samples=num_samples
                )

        else:
            raise ValueError(f"Unknown coordinate covariance sampling method: {method}")

        variants_list.append(
            VariantCoordinates.from_kwargs(
                index=np.full(len(samples), i),
                sample=coordinates.from_kwargs(
                    origin=qv.concatenate(
                        [coordinate_i.origin for i in range(len(samples))]
                    ),
                    time=qv.concatenate(
                        [coordinate_i.time for i in range(len(samples))]
                    ),
                    frame=coordinate_i.frame,
                    **{dim: samples[:, i] for i, dim in enumerate(dimensions)},
                ),
                weight=W,
                weight_cov=W_cov,
            )
        )

    return qv.concatenate(variants_list)



import numpy as np
import pytest

from ...time import Timestamp
from ..cartesian import CartesianCoordinates
from ..cometary import CometaryCoordinates
from ..covariances import CoordinateCovariances
from ..keplerian import KeplerianCoordinates
from ..origin import Origin, OriginCodes
from ..spherical import SphericalCoordinates
from ..transform import transform_coordinates


@pytest.mark.parametrize(
    "representation",
    [SphericalCoordinates, KeplerianCoordinates, CometaryCoordinates],
    ids=lambda x: f"to={x.__name__},",
)
@pytest.mark.parametrize(
    "frame", ["equatorial", "ecliptic"], ids=lambda x: f"frame={x},"
)
@pytest.mark.parametrize(
    "origin",
    [OriginCodes.SUN, OriginCodes.SOLAR_SYSTEM_BARYCENTER],
    ids=lambda x: f"origin={x.name},",
)
@pytest.mark.benchmark(group="coord_transforms")
def test_benchmark_transform_cartesian_coordinates(
    benchmark, representation, frame, origin
):
    if origin == OriginCodes.SOLAR_SYSTEM_BARYCENTER:
        pytest.skip("barycenter transform not yet implemented")

    if frame == "ecliptic":
        frame_in = "equatorial"
    else:
        frame_in = "ecliptic"

    from_coords = CartesianCoordinates.from_kwargs(
        x=np.array([1]),
        y=np.array([1]),
        z=np.array([1]),
        vx=np.array([1]),
        vy=np.array([1]),
        vz=np.array([1]),
        time=Timestamp.from_mjd([50000]),
        origin=Origin.from_kwargs(code=["SUN"]),
        frame=frame_in,
    )
    benchmark(
        transform_coordinates,
        from_coords,
        representation_out=representation,
        frame_out=frame,
        origin_out=origin,
    )


@pytest.mark.benchmark(group="coordinate_covariances")
def test_benchmark_CoordinateCovariances_to_matrix(benchmark):

    covariances_filled = [np.random.random(36) for _ in range(500)]
    covariances_missing = [None for _ in range(500)]
    coordinate_covariances = CoordinateCovariances.from_kwargs(
        values=covariances_filled + covariances_missing
    )
    benchmark(coordinate_covariances.to_matrix)


@pytest.mark.benchmark(group="coordinate_covariances")
def test_benchmark_CoordinateCovariances_from_matrix(benchmark):

    covariances = np.random.random((1000, 6, 6))
    covariances[500:, :, :] = np.nan
    benchmark(CoordinateCovariances.from_matrix, covariances)



from ...utils.helpers.orbits import make_real_orbits


def test_benchmark_iterate_real_orbits(benchmark):
    orbits = make_real_orbits(27)

    def noop_iterate(iterator):
        for _ in iterator:
            pass

    benchmark(noop_iterate, orbits)



from typing import Tuple

import jax.numpy as jnp
import numpy as np
from jax import config, jit, lax

from ..constants import Constants as c
from .stumpff import calc_stumpff

config.update("jax_enable_x64", True)

MU = c.MU


@jit
def calc_chi(
    r: jnp.ndarray,
    v: jnp.ndarray,
    dt: float,
    mu: float = MU,
    max_iter: int = 100,
    tol: float = 1e-16,
) -> Tuple[
    np.float64, np.float64, np.float64, np.float64, np.float64, np.float64, np.float64
]:
    """
    Calculate universal anomaly chi using Newton-Raphson.

    Parameters
    ----------
    r : `~jax.numpy.ndarray` (3)
        Position vector in au.
    v : `~jax.numpy.ndarray` (3)
        Velocity vector in au per day.
    dt : float
        Time from epoch to which calculate chi in units of decimal days.
    mu : float
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will return the value of the universal anomaly at the last iteration.
    tol : float
        Numerical tolerance to which to compute chi using the Newtown-Raphson
        method.

    Returns
    -------
    chi : float
        Universal anomaly.
    c0, c1, c2, c3, c4, c5 : 6 x float
        First six Stumpff functions.

    References
    ----------
    [1] Curtis, H. D. (2014). Orbital Mechanics for Engineering Students. 3rd ed.,
        Elsevier Ltd. ISBN-13: 978-0080977478
    """
    v_mag = jnp.linalg.norm(v)
    r_mag = jnp.linalg.norm(r)
    rv_mag = jnp.dot(r, v) / r_mag
    sqrt_mu = jnp.sqrt(mu)

    # Equations 3.48 and 3.50 in Curtis (2014) [1]
    alpha = -(v_mag**2) / mu + 2 / r_mag

    # Equation 3.66 in Curtis (2014) [1]
    chi = sqrt_mu * jnp.abs(alpha) * dt

    ratio = 1e15
    iterations = 0
    # Define parameters array (arguments that will be changing as
    # the while loop iterates):
    # chi, c0, c1, c2, c3, c4, c5, ratio, iterations
    p = [chi, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ratio, iterations]

    # Define while loop body function
    @jit
    def _chi_newton_raphson(p):

        chi = p[0]
        ratio = p[-2]
        iterations = p[-1]

        chi2 = chi**2
        psi = alpha * chi2
        c0, c1, c2, c3, c4, c5 = calc_stumpff(psi)

        # Newton-Raphson
        # Equation 3.65 in Curtis (2014) [1]
        f = (
            r_mag * rv_mag / sqrt_mu * chi2 * c2
            + (1 - alpha * r_mag) * chi**3 * c3
            + r_mag * chi
            - sqrt_mu * dt
        )
        fp = (
            r_mag * rv_mag / sqrt_mu * chi * (1 - alpha * chi2 * c3)
            + (1 - alpha * r_mag) * chi2 * c2
            + r_mag
        )

        ratio = f / fp
        chi -= ratio
        iterations += 1

        p[0] = chi
        p[1] = c0
        p[2] = c1
        p[3] = c2
        p[4] = c3
        p[5] = c4
        p[6] = c5
        p[7] = ratio
        p[8] = iterations
        return p

    # Define while loop condition function
    @jit
    def _while_condition(p):
        ratio = p[-2]
        iterations = p[-1]
        return (jnp.abs(ratio) > tol) & (iterations <= max_iter)

    p = lax.while_loop(_while_condition, _chi_newton_raphson, p)
    chi = p[0]
    c0 = p[1]
    c1 = p[2]
    c2 = p[3]
    c3 = p[4]
    c4 = p[5]
    c5 = p[6]

    return chi, c0, c1, c2, c3, c4, c5



from typing import List, Union

import numpy.typing as npt
import pandas as pd
from astroquery.jplhorizons import Horizons

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.cometary import CometaryCoordinates
from ...coordinates.keplerian import KeplerianCoordinates
from ...coordinates.origin import Origin
from ...observers import Observers
from ...time import Timestamp
from ..orbits import Orbits


def _get_horizons_vectors(
    object_ids: Union[List, npt.ArrayLike],
    times: Timestamp,
    location: str = "@sun",
    id_type: str = "smallbody",
    aberrations: str = "geometric",
    refplane: str = "ecliptic",
) -> pd.DataFrame:
    """
    Query JPL Horizons (through astroquery) for an object's
    state vectors at the given times.

    Parameters
    ----------
    object_ids : Union[List, `~numpy.ndarray`] (N)
        Object IDs / designations recognizable by HORIZONS.
    times : Timestamp (M)
        Time at which to gather state vectors.
    location : str, optional
        Location of the origin typically a NAIF code.
        ('0' or '@ssb' for solar system barycenter, '10' or '@sun' for heliocenter)
        [Default = '@sun']
    id_type : {'majorbody', 'smallbody', 'designation',
               'name', 'asteroid_name', 'comet_name', 'id'}
        ID type, Horizons will find closest match under any given type.
    aberrations : {'geometric', 'astrometric', 'apparent'}
        Adjust state for one of three different aberrations.
    refplane : {'ecliptic', 'earth'}
        Reference plane for state vectors.

    Returns
    -------
    vectors : `~pandas.DataFrame`
        Dataframe containing the full cartesian state, r, r_rate, delta, delta_rate and light time
        of the object at each time.
    """
    dfs = []
    for i, obj_id in enumerate(object_ids):
        obj = Horizons(
            id=obj_id,
            epochs=times.rescale("tdb").mjd().to_numpy(zero_copy_only=False),
            location=location,
            id_type=id_type,
        )
        vectors = obj.vectors(
            refplane=refplane, aberrations=aberrations, cache=False
        ).to_pandas()
        vectors.insert(0, "orbit_id", f"{i:05d}")
        dfs.append(vectors)

    vectors = pd.concat(dfs, ignore_index=True)
    vectors.sort_values(
        by=["orbit_id", "datetime_jd"],
        inplace=True,
        ignore_index=True,
    )
    return vectors


def _get_horizons_elements(
    object_ids: Union[List, npt.ArrayLike],
    times: Timestamp,
    location: str = "@sun",
    id_type: str = "smallbody",
    refplane: str = "ecliptic",
) -> pd.DataFrame:
    """
    Query JPL Horizons (through astroquery) for an object's
    elements at the given times.

    Parameters
    ----------
    object_ids : Union[List, `~numpy.ndarray`] (N)
        Object IDs / designations recognizable by HORIZONS.
    times : Timestamp
        Time at which to gather state vectors.
    location : str, optional
        Location of the origin typically a NAIF code.
        ('0' or '@ssb' for solar system barycenter, '10' or '@sun' for heliocenter)
        [Default = '@sun']
    id_type : {'majorbody', 'smallbody', 'designation',
               'name', 'asteroid_name', 'comet_name', 'id'}
        ID type, Horizons will find closest match under any given type.
    refplane : {'ecliptic', 'earth'}
        Reference plane for orbital elements.

    Returns
    -------
    elements : `~pandas.DataFrame`
        Dataframe containing the full cartesian state, r, r_rate, delta, delta_rate and light time
        of the object at each time.
    """
    dfs = []
    for i, obj_id in enumerate(object_ids):
        obj = Horizons(
            id=obj_id,
            epochs=times.rescale("tdb").mjd().to_numpy(zero_copy_only=False),
            location=location,
            id_type=id_type,
        )
        elements = obj.elements(
            refsystem="J2000", refplane=refplane, tp_type="absolute", cache=False
        ).to_pandas()
        elements.insert(0, "orbit_id", f"{i:05d}")
        dfs.append(elements)

    elements = pd.concat(dfs, ignore_index=True)
    elements.sort_values(
        by=["orbit_id", "datetime_jd"],
        inplace=True,
        ignore_index=True,
    )
    return elements


def _get_horizons_ephemeris(
    object_ids: Union[List, npt.ArrayLike],
    times: Timestamp,
    location: str,
    id_type: str = "smallbody",
) -> pd.DataFrame:
    """
    Query JPL Horizons (through astroquery) for an object's
    predicted ephemeris as seen from a given location at the given times.

    Parameters
    ----------
    object_ids : Union[List, `~numpy.ndarray`] (N)
        Object IDs / designations recognizable by HORIZONS.
    times : Timestamp
        Time at which to gather state vectors.
    location : str, optional
        Location of the origin typically a NAIF code or MPC observatory code
    id_type : {'majorbody', 'smallbody', 'designation',
               'name', 'asteroid_name', 'comet_name', 'id'}
        ID type, Horizons will find closest match under any given type.

    Returns
    -------
    ephemeris : `~pandas.DataFrame`
        Dataframe containing the predicted ephemerides of the given objects
        as seen from the observer location at the given times.
    """
    dfs = []
    for i, obj_id in enumerate(object_ids):
        obj = Horizons(
            id=obj_id,
            epochs=times.rescale("utc").mjd().to_numpy(zero_copy_only=False),
            location=location,
            id_type=id_type,
        )
        ephemeris = obj.ephemerides(
            # RA, DEC, r, r_rate, delta, delta_rate, lighttime
            # quantities="1, 2, 19, 20, 21",
            extra_precision=True,
            cache=False,
        ).to_pandas()
        ephemeris.insert(0, "orbit_id", f"{i:05d}")
        ephemeris.insert(2, "mjd_utc", times.utc.mjd)
        ephemeris.insert(3, "observatory_code", location)

        dfs.append(ephemeris)

    ephemeris = pd.concat(dfs)
    ephemeris.sort_values(
        by=["orbit_id", "datetime_jd", "observatory_code"],
        inplace=True,
        ignore_index=True,
    )
    return ephemeris


def query_horizons_ephemeris(
    object_ids: Union[List, npt.ArrayLike], observers: Observers
) -> pd.DataFrame:
    """
    Query JPL Horizons (through astroquery) for an object's predicted ephemeris
    as seen from a given location at the given times.

    Parameters
    ----------
    object_ids : Union[List, `~numpy.ndarray`] (N)
        Object IDs / designations recognizable by HORIZONS.
    observers : `~adam_core.observers.observers.Observers`
        Observers object containing the location and times
        of the observers.

    Returns
    -------
    ephemeris : `~pandas.DataFrame`
        Dataframe containing the predicted ephemerides of the given objects
        as seen from the observer location at the given times.
    """
    dfs = []
    for observatory_code, observers_i in observers.iterate_codes():
        ephemeris = _get_horizons_ephemeris(
            object_ids,
            observers_i.coordinates.time,
            observatory_code,
        )
        dfs.append(ephemeris)

    ephemeris = pd.concat(dfs, ignore_index=True)
    ephemeris.sort_values(
        by=["orbit_id", "datetime_jd", "observatory_code"],
        inplace=True,
        ignore_index=True,
    )
    return ephemeris


def query_horizons(
    object_ids: Union[List, npt.ArrayLike],
    times: Timestamp,
    coordinate_type: str = "cartesian",
    location: str = "@sun",
    id_type: str = "smallbody",
    aberrations: str = "geometric",
) -> Orbits:
    """
    Query JPL Horizons (through astroquery) for an object's state vectors or elements at the given times.

    Parameters
    ----------
    object_ids : npt.ArrayLike (N)
        Object IDs / designations recognizable by HORIZONS.
    times : Timestamp (M)
        Time at which to gather state vectors.
    coordinate_type : {'cartesian', 'keplerian', 'cometary'}
        Type of orbital elements to return.
    location : str, optional
        Location of the origin typically a NAIF code.
        ('0' or '@ssb' for solar system barycenter, '10' or '@sun' for heliocenter)
        [Default = '@sun']
    id_type : {'majorbody', 'smallbody', 'designation',
                'name', 'asteroid_name', 'comet_name', 'id'}
        ID type, Horizons will find closest match under any given type.
    aberrations : {'geometric', 'astrometric', 'apparent'}
        Adjust state for one of three different aberrations.

    Returns
    -------
    orbits : `~adam_core.orbits.orbits.Orbits`
        Orbits object containing the state vectors or elements of the object at each time.
    """
    if coordinate_type == "cartesian":
        vectors = _get_horizons_vectors(
            object_ids,
            times,
            location=location,
            id_type=id_type,
            aberrations=aberrations,
        )

        times = Timestamp.from_jd(vectors["datetime_jd"].values, scale="tdb")
        origin = Origin.from_kwargs(code=["SUN" for i in range(len(times))])
        frame = "ecliptic"
        coordinates = CartesianCoordinates.from_kwargs(
            time=times,
            x=vectors["x"].values,
            y=vectors["y"].values,
            z=vectors["z"].values,
            vx=vectors["vx"].values,
            vy=vectors["vy"].values,
            vz=vectors["vz"].values,
            origin=origin,
            frame=frame,
        )
        orbit_id = vectors["orbit_id"].values
        object_id = vectors["targetname"].values

        return Orbits.from_kwargs(
            orbit_id=orbit_id, object_id=object_id, coordinates=coordinates
        )

    elif coordinate_type == "keplerian":
        elements = _get_horizons_elements(
            object_ids,
            times,
            location=location,
            id_type=id_type,
        )

        times = Timestamp.from_jd(
            elements["datetime_jd"].values,
            scale="tdb",
        )
        origin = Origin.from_kwargs(code=["SUN" for i in range(len(times))])
        frame = "ecliptic"
        coordinates = KeplerianCoordinates(
            time=times,
            a=elements["a"].values,
            e=elements["e"].values,
            i=elements["incl"].values,
            raan=elements["Omega"].values,
            ap=elements["w"].values,
            M=elements["M"].values,
            origin=origin,
            frame=frame,
        )
        orbit_id = elements["orbit_id"].values
        object_id = elements["targetname"].values

        return Orbits.from_kwargs(
            orbit_id=orbit_id,
            object_id=object_id,
            coordinates=coordinates.to_cartesian(),
        )

    elif coordinate_type == "cometary":
        elements = _get_horizons_elements(
            object_ids,
            times,
            location=location,
            id_type=id_type,
        )

        tp = Timestamp.from_jd(elements["Tp_jd"].values, scale="tdb")
        times = Timestamp.from_jd(elements["datetime_jd"].values, scale="tdb")
        origin = Origin.from_kwargs(code=["SUN" for i in range(len(times))])
        frame = "ecliptic"
        coordinates = CometaryCoordinates.from_kwargs(
            time=times,
            q=elements["q"].values,
            e=elements["e"].values,
            i=elements["incl"].values,
            raan=elements["Omega"].values,
            ap=elements["w"].values,
            tp=tp.mjd(),
            origin=origin,
            frame=frame,
        )
        orbit_id = elements["orbit_id"].values
        object_id = elements["targetname"].values

        return Orbits.from_kwargs(
            orbit_id=orbit_id,
            object_id=object_id,
            coordinates=coordinates.to_cartesian(),
        )

    else:
        err = "coordinate_type should be one of {'cartesian', 'keplerian', 'cometary'}"
        raise ValueError(err)



import numpy as np
import numpy.testing as npt

from ..transform import (
    _cartesian_to_spherical,
    _spherical_to_cartesian,
    cartesian_to_spherical,
    spherical_to_cartesian,
)

X_AXIS_CARTESIAN = np.array([1.0, 0.0, 0.0, 0.0, 0.0, 0.0])
Y_AXIS_CARTESIAN = np.array([0.0, 1.0, 0.0, 0.0, 0.0, 0.0])
Z_AXIS_CARTESIAN = np.array([0.0, 0.0, 1.0, 0.0, 0.0, 0.0])
NEGATIVE_X_AXIS_CARTESIAN = np.array([-1.0, 0.0, 0.0, 0.0, 0.0, 0.0])
NEGATIVE_Y_AXIS_CARTESIAN = np.array([0.0, -1.0, 0.0, 0.0, 0.0, 0.0])
NEGATIVE_Z_AXIS_CARTESIAN = np.array([0.0, 0.0, -1.0, 0.0, 0.0, 0.0])
CARTESIAN_AXES = [
    X_AXIS_CARTESIAN,
    Y_AXIS_CARTESIAN,
    Z_AXIS_CARTESIAN,
    NEGATIVE_X_AXIS_CARTESIAN,
    NEGATIVE_Y_AXIS_CARTESIAN,
    NEGATIVE_Z_AXIS_CARTESIAN,
]
CARTESIAN_AXES = np.vstack(CARTESIAN_AXES)

X_AXIS_SPHERICAL = np.array([1.0, 0.0, 0.0, 0.0, 0.0, 0.0])
Y_AXIS_SPHERICAL = np.array([1.0, 90.0, 0.0, 0.0, 0.0, 0.0])
Z_AXIS_SPHERICAL = np.array([1.0, 0.0, 90.0, 0.0, 0.0, 0.0])
NEGATIVE_X_AXIS_SPHERICAL = np.array([1.0, 180.0, 0.0, 0.0, 0.0, 0.0])
NEGATIVE_Y_AXIS_SPHERICAL = np.array([1.0, 270.0, 0.0, 0.0, 0.0, 0.0])
NEGATIVE_Z_AXIS_SPHERICAL = np.array([1.0, 0.0, -90.0, 0.0, 0.0, 0.0])
SPHERICAL_AXES = [
    X_AXIS_SPHERICAL,
    Y_AXIS_SPHERICAL,
    Z_AXIS_SPHERICAL,
    NEGATIVE_X_AXIS_SPHERICAL,
    NEGATIVE_Y_AXIS_SPHERICAL,
    NEGATIVE_Z_AXIS_SPHERICAL,
]
SPHERICAL_AXES = np.vstack(SPHERICAL_AXES)


RELATIVE_TOLERANCE = 0.0
ABSOLUTE_TOLERANCE = 1e-15


def test__cartesian_to_spherical():
    # Test _cartesian_to_spherical correctly converts cartesian axes to their equivalent in spherical axes
    # Note velocity conversions are not explicitly tested here
    for cartesian_axis, spherical_axis in zip(CARTESIAN_AXES, SPHERICAL_AXES):
        npt.assert_allclose(
            _cartesian_to_spherical(cartesian_axis),
            spherical_axis,
            rtol=RELATIVE_TOLERANCE,
            atol=ABSOLUTE_TOLERANCE,
        )


def test_cartesian_to_spherical():
    # Test cartesian_to_spherical (vmapped conversion) correctly converts cartesian axes
    # to their equivalent in spherical axes
    # Note velocity conversions are not explicitly tested here
    npt.assert_allclose(
        cartesian_to_spherical(CARTESIAN_AXES),
        SPHERICAL_AXES,
        rtol=RELATIVE_TOLERANCE,
        atol=ABSOLUTE_TOLERANCE,
    )


def test__spherical_to_cartesian():
    # Test _spherical_to_cartesian correctly converts spherical axes to their equivalent in cartesian axes
    # Note velocity conversions are not explicitly tested here
    for cartesian_axis, spherical_axis in zip(CARTESIAN_AXES, SPHERICAL_AXES):
        npt.assert_allclose(
            _spherical_to_cartesian(spherical_axis),
            cartesian_axis,
            rtol=RELATIVE_TOLERANCE,
            atol=ABSOLUTE_TOLERANCE,
        )


def test_spherical_to_cartesian():
    # Test spherical_to_cartesian (vmapped conversion) correctly converts spherical axes
    # to their equivalent in cartesian axes
    # Note velocity conversions are not explicitly tested here
    npt.assert_allclose(
        spherical_to_cartesian(SPHERICAL_AXES),
        CARTESIAN_AXES,
        rtol=RELATIVE_TOLERANCE,
        atol=ABSOLUTE_TOLERANCE,
    )



from __future__ import annotations

import numpy as np
import quivr as qv

from ..time import Timestamp
from . import cartesian, cometary, spherical
from .covariances import CoordinateCovariances, transform_covariances_jacobian
from .origin import Origin

__all__ = [
    "KeplerianCoordinates",
]


class KeplerianCoordinates(qv.Table):

    a = qv.Float64Column()
    e = qv.Float64Column()
    i = qv.Float64Column()
    raan = qv.Float64Column()
    ap = qv.Float64Column()
    M = qv.Float64Column()
    time = Timestamp.as_column(nullable=True)
    covariance = CoordinateCovariances.as_column(nullable=True)
    origin = Origin.as_column()
    frame = qv.StringAttribute(default="unspecified")

    @property
    def values(self) -> np.ndarray:
        return np.array(self.table.select(["a", "e", "i", "raan", "ap", "M"]))

    @property
    def sigma_a(self) -> np.ndarray:
        """
        1-sigma uncertainty in semi-major axis.
        """
        return self.covariance.sigmas[:, 0]

    @property
    def sigma_e(self) -> np.ndarray:
        """
        1-sigma uncertainty in eccentricity.
        """
        return self.covariance.sigmas[:, 1]

    @property
    def sigma_i(self) -> np.ndarray:
        """
        1-sigma uncertainty in inclination.
        """
        return self.covariance.sigmas[:, 2]

    @property
    def sigma_raan(self) -> np.ndarray:
        """
        1-sigma uncertainty in right ascension of the ascending node.
        """
        return self.covariance.sigmas[:, 3]

    @property
    def sigma_ap(self) -> np.ndarray:
        """
        1-sigma uncertainty in argument of periapsis.
        """
        return self.covariance.sigmas[:, 4]

    @property
    def sigma_M(self) -> np.ndarray:
        """
        1-sigma uncertainty in mean anomaly.
        """
        return self.covariance.sigmas[:, 5]

    @property
    def q(self) -> np.ndarray:
        """
        Periapsis distance.
        """
        from ..dynamics.kepler import calc_periapsis_distance

        return np.array(calc_periapsis_distance(self.a.to_numpy(), self.e.to_numpy()))

    @q.setter
    def q(self, value):
        err = (
            "Cannot set periapsis distance (q) as it is"
            " derived from semi-major axis (a) and eccentricity (e)."
        )
        raise ValueError(err)

    @q.deleter
    def q(self):
        err = (
            "Cannot delete periapsis distance (q) as it is"
            " derived from semi-major axis (a) and eccentricity (e)."
        )
        raise ValueError(err)

    @property
    def Q(self) -> np.ndarray:
        """
        Apoapsis distance.
        """
        from ..dynamics.kepler import calc_apoapsis_distance

        return np.array(calc_apoapsis_distance(self.a.to_numpy(), self.e.to_numpy()))

    @Q.setter
    def Q(self, value):
        err = (
            "Cannot set apoapsis distance (Q) as it is"
            " derived from semi-major axis (a) and eccentricity (e)."
        )
        raise ValueError(err)

    @Q.deleter
    def Q(self):
        err = (
            "Cannot delete apoapsis distance (Q) as it is"
            " derived from semi-major axis (a) and eccentricity (e)."
        )
        raise ValueError(err)

    @property
    def p(self) -> np.ndarray:
        """
        Semi-latus rectum.
        """
        from ..dynamics.kepler import calc_semi_latus_rectum

        return np.array(calc_semi_latus_rectum(self.a.to_numpy(), self.e.to_numpy()))

    @p.setter
    def p(self, value):
        err = (
            "Cannot set semi-latus rectum (p) as it is"
            " derived from semi-major axis (a) and eccentricity (e)."
        )
        raise ValueError(err)

    @p.deleter
    def p(self):
        err = (
            "Cannot delete semi-latus rectum (p) as it is"
            " derived from semi-major axis (a) and eccentricity (e)."
        )
        raise ValueError(err)

    @property
    def P(self) -> np.ndarray:
        """
        Period.
        """
        from ..dynamics.kepler import calc_period

        return np.array(calc_period(self.a.to_numpy(), self.origin.mu()))

    @P.setter
    def P(self, value):
        err = (
            "Cannot set period (P) as it is"
            " derived from semi-major axis (a) and gravitational parameter (mu)."
        )
        raise ValueError(err)

    @P.deleter
    def P(self):
        err = (
            "Cannot delete period (P) as it is"
            " derived from semi-major axis (a) and gravitational parameter (mu)."
        )
        raise ValueError(err)

    @property
    def n(self):
        """
        Mean motion in degrees.
        """
        from ..dynamics.kepler import calc_mean_motion

        return np.degrees(
            np.array(calc_mean_motion(self.a.to_numpy(), self.origin.mu()))
        )

    @n.setter
    def n(self, value):
        err = (
            "Cannot set mean motion (n) as it is"
            " derived from semi-major axis (a) and gravitational parameter (mu)."
        )
        raise ValueError(err)

    @n.deleter
    def n(self):
        err = (
            "Cannot delete mean motion (n) as it is"
            " derived from semi-major axis (a) and gravitational parameter (mu)."
        )
        raise ValueError(err)

    def to_cartesian(self) -> cartesian.CartesianCoordinates:
        from .transform import _keplerian_to_cartesian_a, keplerian_to_cartesian

        # Extract gravitational parameter from origin
        mu = self.origin.mu()

        coords_cartesian = keplerian_to_cartesian(
            self.values,
            mu=mu,
            max_iter=1000,
            tol=1e-15,
        )
        coords_cartesian = np.array(coords_cartesian)

        if not self.covariance.is_all_nan():
            covariances_keplerian = self.covariance.to_matrix()
            covariances_cartesian = transform_covariances_jacobian(
                self.values,
                covariances_keplerian,
                _keplerian_to_cartesian_a,
                in_axes=(0, 0, None, None),
                out_axes=0,
                mu=mu,
                max_iter=1000,
                tol=1e-15,
            )
        else:
            covariances_cartesian = np.empty(
                (len(coords_cartesian), 6, 6), dtype=np.float64
            )
            covariances_cartesian.fill(np.nan)

        covariances_cartesian = CoordinateCovariances.from_matrix(covariances_cartesian)
        coords = cartesian.CartesianCoordinates.from_kwargs(
            x=coords_cartesian[:, 0],
            y=coords_cartesian[:, 1],
            z=coords_cartesian[:, 2],
            vx=coords_cartesian[:, 3],
            vy=coords_cartesian[:, 4],
            vz=coords_cartesian[:, 5],
            time=self.time,
            covariance=covariances_cartesian,
            origin=self.origin,
            frame=self.frame,
        )

        return coords

    @classmethod
    def from_cartesian(cls, cartesian: cartesian.CartesianCoordinates):
        from .transform import _cartesian_to_keplerian6, cartesian_to_keplerian

        # Extract gravitational parameter from origin
        mu = cartesian.origin.mu()

        coords_keplerian = cartesian_to_keplerian(
            cartesian.values,
            cartesian.time.to_numpy(),
            mu=mu,
        )
        coords_keplerian = np.array(coords_keplerian)

        if not cartesian.covariance.is_all_nan():
            cartesian_covariances = cartesian.covariance.to_matrix()
            covariances_keplerian = transform_covariances_jacobian(
                cartesian.values,
                cartesian_covariances,
                _cartesian_to_keplerian6,
                in_axes=(0, 0, 0),
                out_axes=0,
                t0=cartesian.time.to_numpy(),
                mu=mu,
            )
        else:
            covariances_keplerian = np.empty(
                (len(coords_keplerian), 6, 6), dtype=np.float64
            )
            covariances_keplerian.fill(np.nan)

        covariances_keplerian = CoordinateCovariances.from_matrix(covariances_keplerian)
        coords = cls.from_kwargs(
            a=coords_keplerian[:, 0],
            e=coords_keplerian[:, 4],
            i=coords_keplerian[:, 5],
            raan=coords_keplerian[:, 6],
            ap=coords_keplerian[:, 7],
            M=coords_keplerian[:, 8],
            time=cartesian.time,
            covariance=covariances_keplerian,
            origin=cartesian.origin,
            frame=cartesian.frame,
        )
        return coords

    def to_cometary(self) -> cometary.CometaryCoordinates:
        return cometary.CometaryCoordinates.from_cartesian(self.to_cartesian())

    @classmethod
    def from_cometary(
        cls, cometary_coordinates: cometary.CometaryCoordinates
    ) -> KeplerianCoordinates:
        return cls.from_cartesian(cometary_coordinates.to_cartesian())

    def to_spherical(self) -> spherical.SphericalCoordinates:
        return spherical.SphericalCoordinates.from_cartesian(self.to_cartesian())

    @classmethod
    def from_spherical(
        cls, spherical_coordinates: spherical.SphericalCoordinates
    ) -> KeplerianCoordinates:
        return cls.from_cartesian(spherical_coordinates.to_cartesian())



import numpy as np

from ...utils.helpers.orbits import make_real_orbits
from ..variants import VariantOrbits


def test_VariantOrbits():

    # Get a sample of real orbits
    orbits = make_real_orbits(10)

    # Create a variant orbits object (expands the covariance matrix)
    # around the mean state
    variant_orbits = VariantOrbits.create(orbits)

    # For these 10 orbits this will select sigma-points so lets
    # check that the number of sigma-points is correct
    assert len(variant_orbits) == len(orbits) * 13

    # Now lets collapse the sigma-points back and see if we can reconstruct
    # the input covairance matrix
    collapsed_orbits = variant_orbits.collapse(orbits)

    # Check that the covariance matrices are close
    np.testing.assert_allclose(
        collapsed_orbits.coordinates.covariance.to_matrix(),
        orbits.coordinates.covariance.to_matrix(),
        rtol=0,
        atol=1e-14,
    )

    # Check that the orbit ids are the same
    np.testing.assert_equal(
        collapsed_orbits.orbit_id.to_numpy(zero_copy_only=False),
        orbits.orbit_id.to_numpy(zero_copy_only=False),
    )



import os
from typing import List, Literal, Optional

import numpy as np
import pyarrow.compute as pc
import spiceypy as sp
from naif_de440 import de440
from naif_earth_itrf93 import earth_itrf93
from naif_eop_high_prec import eop_high_prec
from naif_eop_historical import eop_historical
from naif_eop_predict import eop_predict
from naif_leapseconds import leapseconds

from ..constants import KM_P_AU, S_P_DAY
from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.origin import Origin, OriginCodes
from ..time import Timestamp

DEFAULT_KERNELS = [
    leapseconds,
    de440,
    eop_predict,
    eop_historical,
    eop_high_prec,
    earth_itrf93,
]


J2000_TDB_JD = 2451545.0


def _jd_tdb_to_et(jd_tdb: np.ndarray) -> np.ndarray:
    """
    Convert TDB-scaled JD times to an ephemeris time (ET) in seconds.

    Parameters
    ----------
    jd_tdb : `~numpy.ndarray` (N)
        Times in JD TDB.

    Returns
    -------
    et : `~numpy.ndarray` (N)
        Times in ET in seconds.
    """
    # Convert to days since J2000 (noon on January 1, 2000)
    days_since_j2000 = jd_tdb - J2000_TDB_JD

    # Convert to seconds since J2000
    # (SPICE format)
    et = days_since_j2000 * S_P_DAY
    return et


def setup_SPICE(kernels: Optional[List[str]] = None, force: bool = False):
    """
    Load SPICE kernels.

    This function checks to see if SPICE has already been initialized for the current process.
    If it has, then it does nothing. If it has not, then it loads the desired kernels into SPICE.
    If force is set to True, then the kernels will be loaded regardless of whether or not SPICE
    has already been initialized. SPICE has a limit on the number of kernels that can be loaded
    at once, so it is recommended to only load the kernels that are needed for the current
    calculation (calling sp.furnsh multiple times will load the same kernel multiple times, which
    will cause an error.)

    The default kernels loaded are those provided by the NAIF data packages:

    - Leapsecond data (`naif-leapseconds <https://pypi.org/project/naif-leapseconds/>`_)
    - DE440 ephemeris data (`naif-de440 <https://pypi.org/project/naif-de440/>`_)
    - Longterm Earth Orientation Parameter Predictions (`naif-eop-predict <https://pypi.org/project/naif-eop-predict/>`_)
    - Historical Earth Orientation Parameters (`naif-eop-historical <https://pypi.org/project/naif-eop-historical/>`_)
    - High Precision Earth Orientation Parameters (`naif-eop-high-prec <https://pypi.org/project/naif-eop-high-prec/>`_)
    - Earth Body-fixed Reference Frame/Body Association (`naif-earth-itrf93 <https://pypi.org/project/naif-earth-itrf93/>`_)

    Parameters
    ----------
    kernels :
        List of SPICE kernels to load into SPICE. If None, then the default kernels will be loaded.

    """
    if kernels is None:
        kernels = DEFAULT_KERNELS

    process_id = os.getpid()
    env_var = f"ADAM_CORE_SPICE_INITIALIZED_{process_id}"
    if env_var in os.environ and not force:
        return

    for kernel in kernels:
        sp.furnsh(kernel)
    os.environ[env_var] = "True"
    return


def get_perturber_state(
    perturber: OriginCodes,
    times: Timestamp,
    frame: Literal["ecliptic", "equatorial"] = "ecliptic",
    origin: OriginCodes = OriginCodes.SUN,
) -> CartesianCoordinates:
    """
    Query the JPL ephemeris files loaded in SPICE for the state vectors of desired perturbers.

    Parameters
    ----------
    perturber : OriginCodes
        The NAIF ID of the perturber.
    times : Timestamp (N)
        Times at which to get state vectors.
    frame : {'equatorial', 'ecliptic'}
        Return perturber state in the equatorial or ecliptic J2000 frames.
    origin :  OriginCodes
        The NAIF ID of the origin.

    Returns
    -------
    states : `~adam_core.coordinates.cartesian.CartesianCoordinates`
        The state vectors of the perturber in the desired frame
        and measured from the desired origin.
    """
    if frame == "ecliptic":
        frame_spice = "ECLIPJ2000"
    elif frame == "equatorial":
        frame_spice = "J2000"
    else:
        err = "frame should be one of {'equatorial', 'ecliptic'}"
        raise ValueError(err)

    # Make sure SPICE is ready to roll
    setup_SPICE()

    # Convert epochs to ET in TDB
    epochs_et = times.rescale("tdb").et()
    unique_epochs_et = epochs_et.unique()
    N = len(times)
    # Get position of the body in km and km/s in the desired frame and measured from the desired origin
    states = np.empty((N, 6), dtype=np.float64)

    for i, epoch in enumerate(unique_epochs_et):
        mask = pc.equal(epochs_et, epoch).to_numpy(False)
        state, lt = sp.spkez(
            perturber.value, epoch.as_py(), frame_spice, "NONE", origin.value
        )
        states[mask, :] = state

    # Convert to AU and AU per day
    states = states / KM_P_AU
    states[:, 3:] = states[:, 3:] * S_P_DAY

    return CartesianCoordinates.from_kwargs(
        time=times,
        x=states[:, 0],
        y=states[:, 1],
        z=states[:, 2],
        vx=states[:, 3],
        vy=states[:, 4],
        vz=states[:, 5],
        frame=frame,
        origin=Origin.from_kwargs(code=[origin.name for i in range(N)]),
    )



import logging
from typing import Callable, Optional, Tuple

import numpy as np
import pyarrow as pa
import quivr as qv
from scipy.linalg import sqrtm
from scipy.stats import multivariate_normal

from .jacobian import calc_jacobian

logger = logging.getLogger(__name__)

COVARIANCE_FILL_VALUE = np.NaN


def sigmas_to_covariances(sigmas: np.ndarray) -> np.ndarray:
    """
    Convert an array of standard deviations to an array of covariance matrices.
    Non-diagonal elements are set to zero.

    Parameters
    ----------
    sigmas : `numpy.ndarray` (N, D)
        Standard deviations for N coordinates in D dimensions.

    Returns
    -------
    covariances : `numpy.ndarray` (N, D, D)
        Covariance matrices for N coordinates in D dimensions.
    """
    D = sigmas.shape[1]
    identity = np.identity(D, dtype=sigmas.dtype)
    covariances = np.einsum("kj,ji->kij", sigmas**2, identity, order="C")
    return covariances


class CoordinateCovariances(qv.Table):
    # TODO: Would be interesting if the dimensionality can be generalized
    #      to D dimensions, so (N, D, D) instead of (N, 6, 6). We would be
    #      able to use this class for the covariance matrices of different
    #      measurments like projections (D = 4) and photometry (D = 1).

    values = qv.LargeListColumn(pa.float64(), nullable=True)
    # When fixed, we should revert to:
    # values = Column(pa.fixed_shape_tensor(pa.float64(), (6, 6)))

    @property
    def sigmas(self):
        cov_diag = np.diagonal(self.to_matrix(), axis1=1, axis2=2)
        sigmas = np.sqrt(cov_diag)
        return sigmas

    def to_matrix(self) -> np.ndarray:
        """
        Return the covariance matrices as a 3D array of shape (N, 6, 6).

        Returns
        -------
        covariances : `numpy.ndarray` (N, 6, 6)
            Covariance matrices for N coordinates in 6 dimensions.
        """
        # return self.values.combine_chunks().to_numpy_ndarray()
        values = self.values.to_numpy(zero_copy_only=False)

        # Try and see if all covariance matrices are None, if so return
        # an array of NaNs.
        try:
            if np.all(values == None):  # noqa: E711
                return np.full((len(self), 6, 6), np.nan)

        except ValueError as e:
            err_str = (
                "The truth value of an array with more than one element is ambiguous."
                " Use a.any() or a.all()"
            )
            if str(e) != err_str:
                raise e

        # Try to stack the values into a 3D array. If this works, then
        # all covariance matrices are the same size and we can return
        # the stacked matrices.
        try:
            cov = np.stack(values).reshape(-1, 6, 6)

        # If not then some of the arrays might be None. Lets loop through
        # the values and fill in the arrays that are missing (None) with NaNs.
        except ValueError as e:
            # If we don't get the error we expect, then raise it.
            if str(e) != "all input arrays must have the same shape":
                raise e
            else:
                for i in range(len(values)):
                    if values[i] is None:
                        values[i] = np.full(36, np.nan)

            # Try stacking again
            cov = np.stack(values).reshape(-1, 6, 6)

        return cov

    @classmethod
    def from_matrix(cls, covariances: np.ndarray) -> "CoordinateCovariances":
        """
        Create a Covariances object from a 3D array of covariance matrices.

        Parameters
        ----------
        covariances : `numpy.ndarray` (N, 6, 6)
            Covariance matrices for N coordinates in 6 dimensions.

        Returns
        -------
        covariances : `Covariances`
            Covariance matrices for N coordinates in 6 dimensions.

        Raises
        ------
        ValueError : If the covariance matrices are not (N, 6, 6)
        """
        # cov = pa.FixedShapeTensorArray.from_numpy_ndarray(covariances)
        if covariances.shape[1:] != (6, 6):
            raise ValueError(
                f"Covariance matrices should have shape (N, 6, 6) but got {covariances.shape}"
            )
        cov = covariances.flatten()
        offsets = np.arange(0, (len(covariances) + 1) * 36, 36, dtype=np.int64)
        return cls.from_kwargs(values=pa.LargeListArray.from_arrays(offsets, cov))

    @classmethod
    def from_sigmas(cls, sigmas: np.ndarray) -> "CoordinateCovariances":
        """
        Create a Covariances object from a 2D array of sigmas.

        Parameters
        ----------
        sigmas : `numpy.ndarray` (N, 6)
            Array of 1-sigma uncertainties for N coordinates in 6
            dimensions.

        Returns
        -------
        covariances : `Covariances`
            Covariance matrices with the diagonal elements set to the
            squares of the input sigmas.
        """
        return cls.from_matrix(sigmas_to_covariances(sigmas))

    @classmethod
    def nulls(cls, length: int) -> "CoordinateCovariances":
        """
        Create a Covariances object with all covariance matrix elements set to NaN.
        Parameters
        ----------
        length : `int`
            Number of coordinates.

        Returns
        -------
        covariances : `CoordinateCovariances`
            Covariance matrices for N coordinates in 6 dimensions.
        """
        return cls.from_kwargs(
            values=pa.ListArray.from_arrays(
                pa.array(np.arange(0, 36 * (length + 1), 36)),
                pa.nulls(36 * length, pa.float64()),
            )
        )

    def is_all_nan(self) -> bool:
        """
        Check if all covariance matrix values are NaN.

        Returns
        -------
        is_all_nan : bool
            True if all covariance matrix elements are NaN, False otherwise.
        """
        return np.all(np.isnan(self.to_matrix()))


def sample_covariance_random(
    mean: np.ndarray,
    cov: np.ndarray,
    num_samples: int = 10000,
    seed: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Sample a multivariate Gaussian distribution with given
    mean and covariances.

    The returned weights will be equal to 1 / num_samples so that
    each sample is equally weighted. Weights are returned from this function
    so its interface is identical to that of
    `~adam_core.coordinates.covariances.sample_covariance_sigma_points`.

    Parameters
    ----------
    mean : `~numpy.ndarray` (D)
        Multivariate mean of the Gaussian distribution.
    cov : `~numpy.ndarray` (D, D)
        Multivariate variance-covariance matrix of the Gaussian distribution.
    num_samples : int, optional
        Number of samples to draw from the distribution.

    Returns
    -------
    samples : `~numpy.ndarray` (num_samples, D)
        The drawn samples row-by-row.
    W: `~numpy.ndarray` (num_samples)
        Weights of the samples.
    W_cov: `~numpy.ndarray` (num_samples)
        Weights of the samples to reconstruct covariance matrix.
    """
    normal = multivariate_normal(mean=mean, cov=cov, allow_singular=True, seed=seed)
    samples = normal.rvs(num_samples)
    W = np.full(num_samples, 1 / num_samples)
    W_cov = np.full(num_samples, 1 / num_samples)
    return samples, W, W_cov


def sample_covariance_sigma_points(
    mean: np.ndarray,
    cov: np.ndarray,
    alpha: float = 1,
    beta: float = 0.0,
    kappa: float = 0.0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Create sigma-point samples of a multivariate Gaussian distribution
    with given mean and covariances.

    Parameters
    ----------
    mean : `~numpy.ndarray` (D)
        Multivariate mean of the Gaussian distribution.
    cov : `~numpy.ndarray` (D, D)
        Multivariate variance-covariance matrix of the Gaussian distribution.
    alpha : float, optional
        Spread of the sigma points between 1e^-2 and 1.
    beta : float, optional
        Prior knowledge of the distribution usually set to 2 for a Gaussian.
    kappa : float, optional
        Secondary scaling parameter usually set to 0.

    Returns
    -------
    sigma_points : `~numpy.ndarray` (2 * D + 1, D)
        Sigma points drawn from the distribution.
    W: `~numpy.ndarray` (2 * D + 1)
        Weights of the sigma points.
    W_cov: `~numpy.ndarray` (2 * D + 1)
        Weights of the sigma points to reconstruct covariance matrix.

    References
    ----------
    [1] Wan, E. A; Van Der Merwe, R. (2000). The unscented Kalman filter for nonlinear estimation.
        Proceedings of the IEEE 2000 Adaptive Systems for Signal Processing,
        Communications, and Control Symposium, 153-158.
        https://doi.org/10.1109/ASSPCC.2000.882463
    """
    # Calculate the dimensionality of the distribution
    D = mean.shape[0]

    # See equation 15 in Wan & Van Der Merwe (2000) [1]
    N = 2 * D + 1
    sigma_points = np.empty((N, D))
    W = np.empty(N)
    W_cov = np.empty(N)

    # Calculate the scaling parameter lambda
    lambd = alpha**2 * (D + kappa) - D

    # First sigma point is the mean
    sigma_points[0] = mean

    # Beta is used to encode prior knowledge about the distribution.
    # If the distribution is a well-constrained Gaussian, beta = 2 is optimal
    # but lets set beta to 0 for now which has the effect of not weighting the mean state
    # with 0 for the covariance matrix. This is generally better for more distributions.
    # Calculate the weights for mean and the covariance matrix
    # Weight are used to reconstruct the mean and covariance matrix from the sigma points
    W[0] = lambd / (D + lambd)
    W_cov[0] = W[0] + (1 - alpha**2 + beta)

    # Take the matrix square root of the scaled covariance matrix.
    # Sometimes you'll see this done with a Cholesky decomposition for speed
    # but sqrtm is sufficiently optimized for this use case and typically provides
    # better results
    L = sqrtm((D + lambd) * cov)

    # Calculate the remaining sigma points
    for i in range(D):
        offset = L[i]
        sigma_points[i + 1] = mean + offset
        sigma_points[i + 1 + D] = mean - offset

    # The weights for the remaining sigma points are the same
    # for the mean and the covariance matrix
    W[1:] = 1 / (2 * (D + lambd))
    W_cov[1:] = 1 / (2 * (D + lambd))

    return sigma_points, W, W_cov


def weighted_mean(samples: np.ndarray, W: np.ndarray) -> np.ndarray:
    """
    Calculate the weighted mean of a set of samples.

    Parameters
    ----------
    samples : `~numpy.ndarray` (N, D)
        Samples drawn from the distribution (these can be randomly drawn
        or sigma points)
    W: `~numpy.ndarray` (N)
        Weights of the samples.

    Returns
    -------
    mean : `~numpy.ndarray` (D)
        Mean calculated from the samples and weights.
    """
    return np.dot(W, samples)


def weighted_covariance(
    mean: np.ndarray, samples: np.ndarray, W_cov: np.ndarray
) -> np.ndarray:
    """
    Calculate a covariance matrix from samples and their corresponding weights.

    Parameters
    ----------
    mean : `~numpy.ndarray` (D)
        Mean calculated from the samples and weights.
        See `~adam_core.coordinates.covariances.weighted_mean`.
    samples : `~numpy.ndarray` (N, D)
        Samples drawn from the distribution (these can be randomly drawn
        or sigma points)
    W_cov: `~numpy.ndarray` (N)
        Weights of the samples to reconstruct covariance matrix.

    Returns
    -------
    cov : `~numpy.ndarray` (D, D)
        Covariance matrix calculated from the samples and weights.
    """
    # Calculate the covariance matrix from the sigma points and weights
    # `~numpy.cov` does not support negative weights so we will calculate
    # the covariance manually
    # cov = np.cov(samples, aweights=W_cov, rowvar=False, bias=True)
    residual = samples - mean
    cov = (W_cov * residual.T) @ residual
    return cov


def transform_covariances_sampling(
    coords: np.ndarray,
    covariances: np.ndarray,
    func: Callable,
    num_samples: int = 100000,
) -> np.ma.masked_array:
    """
    Transform covariance matrices by sampling the transformation function.

    Parameters
    ----------
    coords : `~numpy.ndarray` (N, D)
        Coordinates that correspond to the input covariance matrices.
    covariances : `~numpy.ndarray` (N, D, D)
        Covariance matrices to transform via sampling.
    func : function
        A function that takes coords (N, D) as input and returns the transformed
        coordinates (N, D). See for example: `thor.coordinates.cartesian_to_spherical`
        or `thor.coordinates.cartesian_to_keplerian`.
    num_samples : int, optional
        The number of samples to draw.

    Returns
    -------
    covariances_out : `~numpy.ndarray` (N, D, D)
        Transformed covariance matrices.
    """
    covariances_out = []
    for coord, covariance in zip(coords, covariances):
        samples, W, W_cov = sample_covariance_random(coord, covariance, num_samples)
        samples_converted = func(samples)
        covariances_out.append(np.cov(samples_converted.T))

    covariances_out = np.stack(covariances_out)
    return covariances_out


def transform_covariances_jacobian(
    coords: np.ndarray,
    covariances: np.ndarray,
    _func: Callable,
    **kwargs,
) -> np.ndarray:
    """
    Transform covariance matrices by calculating the Jacobian of the transformation function
    using `~jax.jacfwd`.

    Parameters
    ----------
    coords : `~numpy.ndarray` (N, D)
        Coordinates that correspond to the input covariance matrices.
    covariances : `~numpy.ndarray` (N, D, D)
        Covariance matrices to transform via numerical differentiation.
    _func : function
        A function that takes a single coord (D) as input and return the transformed
        coordinate (D). See for example: `thor.coordinates._cartesian_to_spherical`
        or `thor.coordinates._cartesian_to_keplerian`.

    Returns
    -------
    covariances_out : `~numpy.ndarray` (N, D, D)
        Transformed covariance matrices.
    """
    jacobian = calc_jacobian(coords, _func, **kwargs)
    covariances = jacobian @ covariances @ np.transpose(jacobian, axes=(0, 2, 1))
    return covariances



from copy import deepcopy
from typing import List, Union

import astropy.units
import numpy as np

from .types import Coordinates

# This module is not maintained.
__doctest_skip__ = ["*"]


def _convert_coordinates_units(
    coords: Union[np.ndarray, np.ma.masked_array], units: List, desired_units: List
) -> Union[np.ndarray, np.ma.masked_array]:
    """
    Convert coordinate units to desired units.

    Parameters
    ----------
    coords : `~numpy.ndarray` or `~numpy.ma.masked_array` (N, D)
        Coordinates that need to be converted.
    units : List (D)
        Current units for each coordinate dimension.
    desired_units : List (D)
        Desired units for each coordinate dimension.

    Returns
    -------
    coords_converted : `~numpy.ndarray` or `~numpy.ma.masked_array` (N, D)
        Coordinates converted to the desired coordinate units.

    Raises
    ------
    ValueError : If units or desired_units do not have length D.
    """
    N, D = coords.shape
    coords_converted = coords.copy()

    if (len(units) != D) or (len(desired_units) != D):
        err = f"Length of units or desired_units does not match the number of coordinate dimensions (D={D})"
        raise ValueError(err)

    for i in range(D):
        coords_converted[:, i] = coords[:, i] * units[i].to(desired_units[i])

    return coords_converted


def _convert_covariances_units(
    covariances: Union[np.ndarray, np.ma.masked_array],
    units: np.ndarray,
    desired_units: np.ndarray,
) -> Union[np.ndarray, np.ma.masked_array]:
    """
    Convert covariance units to desired units.

    Parameters
    ----------
    covariances : `~numpy.ndarray` or `~numpy.ma.masked_array` (N, D, D)
        Covariances that need to be converted.
    units : `~numpy.ndarray` (D)
        Current units for each coordinate dimension. Note, these are not the units for
        the elements of the covariance matrices but the actual units of the mean values/states to
        which the covariance matrices belong.
    desired_units : `~numpy.ndarray` (D)
        Desired units for each coordinate dimension.

    Returns
    -------
    covariances_converted : `~numpy.ndarray` or `~numpy.ma.masked_array` (N, D, D)
        Covariances converted to the desired coordinate units.

    Raises
    ------
    ValueError : If units or desired_units do not have length D.
    """
    N, D, D = covariances.shape

    if (len(units) != D) or (len(desired_units) != D):
        err = f"Length of units or desired_units does not match the number of coordinate dimensions (D={D})"
        raise ValueError(err)

    coords_units_2d = units.copy().reshape(1, -1)
    covariance_units = np.dot(coords_units_2d.T, coords_units_2d)

    desired_units_2d = desired_units.copy().reshape(1, -1)
    desired_units = np.dot(desired_units_2d.T, desired_units_2d)

    conversion_matrix = np.zeros((D, D), dtype=np.float64)
    for i in range(D):
        for j in range(D):
            conversion_matrix[i, j] = covariance_units[i, j].to(desired_units[i, j])

    covariances_converted = conversion_matrix * covariances

    return covariances_converted


def convert_coordinates(
    coords: Coordinates,
    desired_units: Union[List[astropy.units.Unit], dict[str, astropy.units.Unit]],
):
    """
    Convert the units in a Coordinates object to desired units.

    Parameters
    ----------
    coords :
        Coordinates that need to be converted.
    desired_units : list, `~numpy.ndarray`, dict
        Desired units for each coordinate dimension expressed as an array-like
        or a dictionary keyed on coordinate dimensions and with values that represent
        the desired units. If a dictionary is passed, then only the coordinate dimensions
        that need to be converted need to be defined.

    Returns
    -------
    coords_converted : `~thor.coordinates.Coordinates` (N, D)
        Coordinates converted to the desired coordinate units.

    Raises
    ------
    ValueError : If units or desired_units do not have length D.
    ValueError : If desired_units is not a list, `~numpy.ndarray`, dict

    Examples
    --------
    >>> import astropy.units
    >>> from adam_core.coordinates import CartesianCoordinates, convert_coordinates
    >>> # By default, x, y, and z are in AU
    >>> coords = CartesianCoordinates.from_kwargs(
    ...    x=[1, 1, 1],
    ...    y=[4, 5, 6],
    ...    z=[7, 8, 9],
    ...    vx=[10, 11, 12],
    ...    vy=[13, 14, 15],
    ...    vz=[16, 17, 18],
    ... )
    >>> desired_units = {
    ...     "x" : astropy.units.km
    ... }
    >>> coordinates_converted = convert_coordinates(coords, desired_units)
    >>> print(coordinates_converted.x.to_pylist())
    [149597870.0, 149597870.0, 149597870.0]
    >>> print(coordinates_converted.y.to_pylist())
    [4.0, 5.0, 6.0]

    """
    N, D = coords.values.shape

    desired_units_ = {}
    if isinstance(desired_units, dict):
        for k, v in coords.units.items():
            if k not in desired_units:
                desired_units_[k] = v
            else:
                desired_units_[k] = desired_units[k]

    elif isinstance(desired_units, (list, np.ndarray)):
        if len(desired_units) != D:
            err = (
                "Length of units or desired_units does not match the"
                f" number of coordinate dimensions (D={D})"
            )
            raise ValueError(err)

        for i, (k, v) in enumerate(coords.units.items()):
            desired_units_[k] = desired_units[i]

    else:
        err = "desired_units should be one of {list, '~numpy.ndarray`, dict}"
        raise ValueError(err)

    # Make a copy of the input coordinates
    coords_converted = deepcopy(coords)
    desired_units_array = np.array(list(desired_units_.values()))
    current_units = np.array(list(coords.units.values()))

    # Modify the underlying coordinates array
    values_converted = _convert_coordinates_units(
        coords.values,
        current_units,
        desired_units_array,
    )
    coords_converted._values = values_converted

    # Modify the underlying covariance matrices
    if coords.covariances is not None:
        covariances_converted = _convert_covariances_units(
            coords.covariances,
            current_units,
            desired_units_array,
        )
        coords_converted._covariances = covariances_converted

    # Update the units dictionary to reflect the new units
    coords_converted._units = desired_units_

    return coords_converted



import warnings
from typing import List, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import pyarrow.compute as pc
from scipy.optimize import least_squares

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.covariances import CoordinateCovariances
from ..coordinates.origin import Origin
from ..coordinates.residuals import Residuals
from ..orbits.orbits import Orbits
from ..propagator.propagator import Propagator
from ..time.time import Timestamp
from .evaluate import OrbitDeterminationObservations, evaluate_orbits
from .fitted_orbits import FittedOrbitMembers, FittedOrbits


def residual_function(
    state_vector: npt.NDArray[np.float64],
    mjd_tdb: float,
    observations: OrbitDeterminationObservations,
    propagator: Propagator,
) -> npt.NDArray[np.float64]:
    """
    Compute the residuals for a given Cartesian orbit with respect to a
    a set of observations.

    Parameters
    ----------
    state_vector : `~numpy.ndarray` (6,)
        State vector of the orbit in Cartesian coordinates.
    mjd_tdb : float
        Time of the state vector in MJD TDB.
    observations : `~thor.orbit_determination.DifferentialCorrectionObservations` (N)
        Observations to compute residuals for.
    propagator : `~thor.propagator.Propagator`
        Propagator to use to generate ephemeris.

    Returns
    -------
    residuals : `~numpy.ndarray` (N,)
        Weighted residual (chi) for each observation.
    """
    # Generate ephemeris and compute residuals
    orbit = Orbits.from_kwargs(
        coordinates=CartesianCoordinates.from_kwargs(
            x=state_vector[0:1],
            y=state_vector[1:2],
            z=state_vector[2:3],
            vx=state_vector[3:4],
            vy=state_vector[4:5],
            vz=state_vector[5:6],
            time=Timestamp.from_mjd([mjd_tdb], scale="tdb"),
            origin=Origin.from_kwargs(code=["SUN"]),
            frame="ecliptic",
        )
    )
    ephemeris = propagator.generate_ephemeris(
        orbit, observations.observers, max_processes=1
    )
    residuals = Residuals.calculate(observations.coordinates, ephemeris.coordinates)

    # We return the "chi2" value from adam_core.coordinates.residuals. This is equivalent to
    # the mahalanobis distance squared (which is just the squared residual normalized by the
    # uncertainty on the observations). This effectively allows us to the weight the
    # observations (and their corresponding residuals) by their uncertainty.
    return np.sqrt(residuals.chi2.to_numpy())


def fit_least_squares(
    orbit: Orbits,
    observations: OrbitDeterminationObservations,
    propagator: Propagator,
    ignore: Optional[List[str]] = None,
    **kwargs,
) -> Tuple[FittedOrbits, FittedOrbitMembers]:
    """
    Differentially correct a fitted orbit using least squares. correct a fitted orbit using least squares.

    Parameters
    ----------
    orbit : `~adam_core.orbits.Orbits` (1)
        Orbit to differentially correct.
    observations : `~adam_core.orbit_determination.DifferentialCorrectionObservations` (N)
        Observations.
    propagator : `~adam_core.propagator.Propagator`
        Propagator to use to generate ephemeris.
    ignore : list of str
        List of observation IDs to ignore when fitting the orbit with least squares.
        These observations will be marked as outliers in the fitted orbit members.
    **kwargs
        Additional keyword arguments to pass to `~scipy.optimize.least_squares`.
        Some of these parameters if not specified will be set to sensible defaults.
            xtol = 1e-12
            ftol = 1e-12
            gtol = 1e-12
            jac = "2-point"
            x_scale = "jac"
            bounds = (-np.inf, np.inf) (for each parameter)

    Returns
    -------
    fitted_orbit : `~adam_core.orbit_determination.FittedOrbits` (1)
        Fitted orbit.
    fitted_orbit_members : `~adam_core.orbit_determination.FittedOrbitMembers` (N)
        Fitted orbit members.
    """
    assert len(orbit) == 1, "Only one orbit can be differentially corrected"

    # TODO: Investigate whether we want to add fitting for the epoch as well
    # Set up least squares problem
    if ignore is not None:
        mask = pc.invert(pc.is_in(observations.id, pa.array(ignore)))
        observations_to_include = observations.apply_mask(mask)
    else:
        observations_to_include = observations

    parameters = 6
    # Extract epoch and state vector from orbit
    epoch = orbit.coordinates.time.mjd().to_numpy()
    state_vector = orbit.coordinates.values[0]
    args = (epoch[0], observations_to_include, propagator)

    # Define some sensible defaults for the least squares fitting procedure
    if "xtol" not in kwargs:
        kwargs["xtol"] = 1e-12
    if "ftol" not in kwargs:
        kwargs["ftol"] = 1e-12
    if "gtol" not in kwargs:
        kwargs["gtol"] = 1e-12
    if "jac" not in kwargs:
        kwargs["jac"] = "2-point"
    if "x_scale" not in kwargs:
        kwargs["x_scale"] = "jac"
    if "bounds" not in kwargs:
        kwargs["bounds"] = (
            np.full(parameters, -np.inf),
            np.full(parameters, np.inf),
        )
    if "args" in kwargs:
        kwargs.pop("args")
        warnings.warn(
            "The args parameter is not supported and will be ignored.",
            category=RuntimeWarning,
        )

    # Run least squares
    solution = least_squares(residual_function, state_vector, args=args, **kwargs)

    # Extract solution state vector and covariance matrix
    mjd_tdb = epoch[0]
    x, y, z, vx, vy, vz = solution.x
    try:
        covariance_matrix = np.linalg.inv(solution.jac.T @ solution.jac)
    except np.linalg.LinAlgError:
        warnings.warn(
            "The covariance matrix could not be computed. The solution may be "
            "unreliable.",
            category=RuntimeWarning,
        )
        covariance_matrix = np.full((6, 6), np.nan)

    # Create orbit with solution state vector and use it to generate ephemeris
    # and calculate the residuals with respect to the observations
    orbit = Orbits.from_kwargs(
        orbit_id=orbit.orbit_id,
        object_id=orbit.object_id,
        coordinates=CartesianCoordinates.from_kwargs(
            x=[x],
            y=[y],
            z=[z],
            vx=[vx],
            vy=[vy],
            vz=[vz],
            time=Timestamp.from_mjd([mjd_tdb], scale="tdb"),
            covariance=CoordinateCovariances.from_matrix(
                covariance_matrix.reshape(1, 6, 6)
            ),
            origin=orbit.coordinates.origin,
            frame=orbit.coordinates.frame,
        ),
    )

    # Evaluate the solution orbit and return it as a fitted orbit and fitted orbit members
    # which contain the residuals with respect to the observations and the overall
    # quality of the fit
    fitted_orbit, fitted_orbit_members = evaluate_orbits(
        orbit,
        observations,
        propagator,
        parameters=parameters,
        ignore=ignore,
    )
    fitted_orbit = (
        fitted_orbit.set_column("iterations", [solution.nfev])
        .set_column("success", [solution.success])
        .set_column("status_code", [solution.status])
    )
    fitted_orbit_members = fitted_orbit_members.set_column(
        "solution", pc.invert(fitted_orbit_members.outlier)
    )

    return fitted_orbit, fitted_orbit_members



# flake8: noqa: F401
from .observations import make_observations
from .orbits import make_real_orbits, make_simple_orbits

__all__ = [
    "make_observations",
    "make_real_orbits",
    "make_simple_orbits",
]



from typing import Tuple

import jax.numpy as jnp
from jax import config, jit

from ..constants import Constants as C
from .chi import calc_chi
from .stumpff import STUMPFF_TYPES

config.update("jax_enable_x64", True)


MU = C.MU
LAGRANGE_TYPES = Tuple[jnp.float64, jnp.float64, jnp.float64, jnp.float64]


@jit
def calc_lagrange_coefficients(
    r: jnp.ndarray,
    v: jnp.ndarray,
    dt: float,
    mu: float = MU,
    max_iter: int = 100,
    tol: float = 1e-16,
) -> Tuple[LAGRANGE_TYPES, STUMPFF_TYPES, jnp.float64]:
    """
    Calculate the exact Lagrange coefficients given an initial state defined at t0,
    and the change in time from t0 to t1 (dt = t1 - t0).

    Parameters
    ----------
    r : `~jax.numpy.ndarray` (3)
        Position vector in au.
    v : `~jax.numpy.ndarray` (3)
        Velocity vector in au per day.
    dt : float
        Time from epoch to which calculate chi in units of decimal days.
    mu : float
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will return the value of the universal anomaly at the last iteration.
    tol : float
        Numerical tolerance to which to compute chi using the Newtown-Raphson
        method.

    Returns
    -------
    lagrange_coeffs : (float x 4)
        f : float
            Langrange f coefficient.
        g : float
            Langrange g coefficient.
        f_dot : float
            Time deriviative of the Langrange f coefficient.
        g_dot : float
            Time deriviative of the Langrange g coefficient.
    stumpff_coeffs : (float x 6)
        First six Stumpff functions (c0, c1, c2, c3, c4, c5)
    chi : float
        Universal anomaly.

    References
    ----------
    [1] Curtis, H. D. (2014). Orbital Mechanics for Engineering Students. 3rd ed.,
        Elsevier Ltd. ISBN-13: 978-0080977478
    """
    sqrt_mu = jnp.sqrt(mu)
    chi, c0, c1, c2, c3, c4, c5 = calc_chi(r, v, dt, mu=mu, max_iter=max_iter, tol=tol)
    stumpff_coeffs = (c0, c1, c2, c3, c4, c5)
    chi2 = chi**2

    r_mag = jnp.linalg.norm(r)
    v_mag = jnp.linalg.norm(v)

    # Equations 3.48 and 3.50 in Curtis (2014) [1]
    alpha = -(v_mag**2) / mu + 2 / r_mag

    # Equations 3.69a and 3.69b in Curtis (2014) [1]
    f = 1 - chi**2 / r_mag * c2
    g = dt - 1 / sqrt_mu * chi**3 * c3

    r_new = f * r + g * v
    r_new_mag = jnp.linalg.norm(r_new)

    # Equations 3.69c and 3.69d in Curtis (2014) [1]
    f_dot = sqrt_mu / (r_mag * r_new_mag) * (alpha * chi**3 * c3 - chi)
    g_dot = 1 - chi2 / r_new_mag * c2

    lagrange_coeffs = (f, g, f_dot, g_dot)

    return lagrange_coeffs, stumpff_coeffs, chi


@jit
def apply_lagrange_coefficients(
    r: jnp.ndarray, v: jnp.ndarray, f: float, g: float, f_dot: float, g_dot: float
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Apply the Lagrange coefficients to r and v.

    Parameters
    ----------
    r : `~jax.numpy.ndarray` (3)
        Position vector in au.
    v : `~jax.numpy.ndarray` (3)
        Velocity vector in au per day.
    f : float
        Langrange f coefficient.
    g : float
        Langrange g coefficient.
    f_dot : float
        Time deriviative of the Langrange f coefficient.
    g_dot : float
        Time deriviative of the Langrange g coefficient.

    Returns
    -------
    r_new : `~jax.numpy.ndarray` (3)
        New position vector in au propagated with the Lagrange coefficients.
    v_new : `~jax.numpy.ndarray` (3)
        New velocity vector in au per day propagated with the Lagrange coefficients.
    """
    r_new = f * r + g * v
    v_new = f_dot * r + g_dot * v

    return r_new, v_new



from __future__ import annotations

import logging

import numpy as np
import numpy.typing as npt
import quivr as qv

from ..time import Timestamp
from . import cometary, keplerian, spherical
from .covariances import CoordinateCovariances
from .origin import Origin

__all__ = ["CartesianCoordinates"]

COVARIANCE_ROTATION_TOLERANCE = 1e-25
logger = logging.getLogger(__name__)


class CartesianCoordinates(qv.Table):
    """Represents coordinates in Cartesian space."""

    #: x coordinates in AU
    x = qv.Float64Column(nullable=True)

    #: y coordinates in AU
    y = qv.Float64Column(nullable=True)

    #: z coordinates in AU
    z = qv.Float64Column(nullable=True)

    #: x velocity in AU/day
    vx = qv.Float64Column(nullable=True)

    #: y velocity in AU/day
    vy = qv.Float64Column(nullable=True)

    #: z velocity in AU/day
    vz = qv.Float64Column(nullable=True)

    #: The instant at which the coordinates are valid
    time = Timestamp.as_column(nullable=True)

    #: Covariance matrix for the x, y, z, vx, vy, and vz values
    covariance = CoordinateCovariances.as_column(nullable=True)

    #: Center of the coordinate system
    origin = Origin.as_column()

    #: Frame of the coordinate system - 'ecliptic' or 'equatorial' or 'unspecified'.
    frame = qv.StringAttribute(default="unspecified")

    @property
    def values(self) -> npt.NDArray[np.float64]:
        """The x, y, z, vx, vy, and vz columns all in one 6-N matrix
        of numpy float64 values. Nulls are converted to NaNs.

        """
        return np.array(self.table.select(["x", "y", "z", "vx", "vy", "vz"]))

    @property
    def r(self) -> npt.NDArray[np.float64]:
        """
        Position vector.
        """
        return np.array(self.table.select(["x", "y", "z"]))

    @property
    def r_mag(self) -> npt.NDArray[np.float64]:
        """
        Magnitude of the position vector.
        """
        return np.linalg.norm(self.r, axis=1)

    @property
    def r_hat(self) -> npt.NDArray[np.float64]:
        """
        Unit vector in the direction of the position vector.
        """
        return self.r / self.r_mag[:, None]

    @property
    def v(self) -> npt.NDArray[np.float64]:
        """
        Velocity vector.
        """
        return np.array(self.table.select(["vx", "vy", "vz"]))

    @property
    def v_mag(self) -> npt.NDArray[np.float64]:
        """
        Magnitude of the velocity vector.
        """
        return np.linalg.norm(self.v, axis=1)

    @property
    def v_hat(self) -> npt.NDArray[np.float64]:
        """
        Unit vector in the direction of the velocity vector.
        """
        return self.v / self.v_mag[:, None]

    @property
    def sigma_x(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the X-coordinate.
        """
        return self.covariance.sigmas[:, 0]

    @property
    def sigma_y(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the Y-coordinate.
        """
        return self.covariance.sigmas[:, 1]

    @property
    def sigma_z(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the Z-coordinate.
        """
        return self.covariance.sigmas[:, 2]

    @property
    def sigma_vx(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the X-coordinate velocity.
        """
        return self.covariance.sigmas[:, 3]

    @property
    def sigma_vy(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the Y-coordinate velocity.
        """
        return self.covariance.sigmas[:, 4]

    @property
    def sigma_vz(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the Z-coordinate velocity.
        """
        return self.covariance.sigmas[:, 5]

    @property
    def sigma_r(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainties in the position vector.
        """
        return self.covariance.sigmas[:, 0:3]

    @property
    def sigma_r_mag(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the position.
        """
        return np.sqrt(np.sum(self.covariance.sigmas[:, 0:3] ** 2, axis=1))

    @property
    def sigma_v(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainties in the velocity vector.
        """
        return self.covariance.sigmas[:, 3:6]

    @property
    def sigma_v_mag(self) -> npt.NDArray[np.float64]:
        """
        1-sigma uncertainty in the velocity vector.
        """
        return np.sqrt(np.sum(self.covariance.sigmas[:, 3:6] ** 2, axis=1))

    @property
    def h(self) -> npt.NDArray[np.float64]:
        """
        Specific angular momentum vector.
        """
        return np.cross(self.r, self.v)

    @property
    def h_mag(self) -> npt.NDArray[np.float64]:
        """
        Magnitude of the specific angular momentum vector.
        """
        return np.linalg.norm(self.h, axis=1)

    def rotate(
        self, rotation_matrix: npt.NDArray[np.float64], frame_out: str
    ) -> CartesianCoordinates:
        """
        Rotate Cartesian coordinates and their covariances by the given rotation matrix.

        Covariance matrices are also rotated. Rotations will sometimes result
        in covariance matrix elements very near zero but not exactly zero. Any
        elements that are smaller than +-1e-25 are rounded down to 0.

        Parameters
        ----------
        matrix:
            6x6 rotation matrix.
        frame_out: str
            Name of the frame to which coordinates are being rotated.

        Returns
        -------
            Rotated Cartesian coordinates and their covariances.
        """
        # Extract coordinate values into a masked array and mask NaNss
        masked_coords = np.ma.masked_array(self.values, fill_value=np.NaN)
        masked_coords.mask = np.isnan(masked_coords.data)

        # Rotate coordinates
        coords_rotated = np.ma.dot(masked_coords, rotation_matrix.T, strict=False)

        # Extract covariances
        masked_covariances = np.ma.masked_array(
            self.covariance.to_matrix(), fill_value=0.0
        )
        masked_covariances.mask = np.isnan(masked_covariances.data)

        # Rotate covariances
        covariances_rotated = (
            rotation_matrix @ masked_covariances.filled() @ rotation_matrix.T
        )
        # Reset the mask to the original mask
        covariances_rotated[masked_covariances.mask] = np.NaN

        # Check if any covariance elements are near zero, if so set them to zero
        near_zero = len(
            covariances_rotated[
                np.abs(covariances_rotated) < COVARIANCE_ROTATION_TOLERANCE
            ]
        )
        if near_zero > 0:
            logger.debug(
                f"{near_zero} covariance elements are within {COVARIANCE_ROTATION_TOLERANCE:.0e}"
                " of zero after rotation, setting these elements to 0."
            )
            covariances_rotated = np.where(
                np.abs(covariances_rotated) < COVARIANCE_ROTATION_TOLERANCE,
                0,
                covariances_rotated,
            )

        coords = self.from_kwargs(
            x=coords_rotated[:, 0],
            y=coords_rotated[:, 1],
            z=coords_rotated[:, 2],
            vx=coords_rotated[:, 3],
            vy=coords_rotated[:, 4],
            vz=coords_rotated[:, 5],
            time=self.time,
            covariance=CoordinateCovariances.from_matrix(covariances_rotated),
            origin=self.origin,
            frame=frame_out,
        )
        return coords

    def translate(
        self, vector: npt.NDArray[np.float64], origin_out: str
    ) -> CartesianCoordinates:
        """
        Translate Cartesian coordinates by the given vector.

        Parameters
        ----------

        vector:
            6x1 or 6xN translation vector. If a 6x1 vector is given, it is applied to all coordinates.
            If an 6xN array of vectors is given, each vector is applied to the corresponding coordinate.
        origin_out:
            Name of the origin to which coordinates are being translated.

        Returns
        -------
            Translated Cartesian coordinates and their covariances.
        """
        N = len(self)
        if vector.shape == (6,):
            vector_ = vector.reshape(1, 6)
        elif vector.shape == (N, 6):
            vector_ = vector
        else:
            raise ValueError(f"Expected vector to have shape (6,) or ({N}, 6).")

        # Extract coordinate values into a masked array and mask NaNss
        masked_coords = np.ma.masked_array(self.values, fill_value=np.NaN)
        masked_coords.mask = np.isnan(masked_coords.data)

        # Translate coordinates
        coords_translated = (masked_coords + vector_).filled()

        coords = self.from_kwargs(
            x=coords_translated[:, 0],
            y=coords_translated[:, 1],
            z=coords_translated[:, 2],
            vx=coords_translated[:, 3],
            vy=coords_translated[:, 4],
            vz=coords_translated[:, 5],
            time=self.time,
            covariance=self.covariance,
            origin=Origin.from_kwargs(code=[origin_out] * len(self)),
            frame=self.frame,
        )
        return coords

    def to_cometary(self) -> cometary.CometaryCoordinates:
        """
        Converts the Cartesian coordinates to the cometary parameterization.
        """
        return cometary.CometaryCoordinates.from_cartesian(self)

    @classmethod
    def from_cometary(
        cls, cometary: cometary.CometaryCoordinates
    ) -> CartesianCoordinates:
        """
        Constructs CartesianCoordinates from the cometary parameterization.
        """
        return cometary.to_cartesian()

    def to_keplerian(self) -> keplerian.KeplerianCoordinates:
        """
        Converts the Cartesian coordinates to the Keplerian parameterization.
        """
        return keplerian.KeplerianCoordinates.from_cartesian(self)

    @classmethod
    def from_keplerian(
        cls, keplerian: keplerian.KeplerianCoordinates
    ) -> CartesianCoordinates:
        """
        Constructs CartesianCoordinates from the Keplerian parameterization.
        """
        return keplerian.to_cartesian()

    def to_spherical(self) -> spherical.SphericalCoordinates:
        """
        Converts the Cartesian coordinates to the spherical parameterization.
        """
        return spherical.SphericalCoordinates.from_cartesian(self)

    @classmethod
    def from_spherical(
        cls, spherical: spherical.SphericalCoordinates
    ) -> CartesianCoordinates:
        """
        Constructs CartesianCoordinates from the spherical parameterization.
        """
        return spherical.to_cartesian()



import numpy as np

__all__ = [
    "KM_P_AU",
    "S_P_DAY",
    "Constants",
    "DE44X",
]

# km in an au
KM_P_AU = 149597870.700
# seconds in a day
S_P_DAY = 86400.0


class _Constants:
    def __init__(self, C=None, MU=None, R_Earth=None, Obliquity=None):
        self.C = C
        self.MU = MU
        self.R_EARTH = R_Earth
        self.OBLIQUITY = Obliquity

        # Transformation matrix from Equatorial J2000 to Ecliptic J2000
        self.TRANSFORM_EQ2EC = np.array(
            [
                [1, 0, 0],
                [0, np.cos(self.OBLIQUITY), np.sin(self.OBLIQUITY)],
                [0, -np.sin(self.OBLIQUITY), np.cos(self.OBLIQUITY)],
            ]
        )

        # Transformation matrix from Ecliptic J2000 to Equatorial J2000
        self.TRANSFORM_EC2EQ = self.TRANSFORM_EQ2EC.T
        return


DE44X_CONSTANTS = {
    # Speed of Light : au / d (299792.458 km / s -- DE430/DE431)
    "C": 299792.458 / KM_P_AU * S_P_DAY,
    # Standard Gravitational Parameter -- Sun :  au**3 / d**2 (0.29591220828411956E-03 -- DE441/DE440)
    "MU": 0.29591220828411956e-03,
    # Earth Equatorial Radius: au (6378.1363 km -- DE431/DE430)
    "R_Earth": 6378.1363 / KM_P_AU,
    # Mean Obliquity at J2000: radians (84381.448 arcseconds -- DE431/DE430)
    "Obliquity": 84381.448 * np.pi / (180.0 * 3600.0),
}
DE44X = Constants = _Constants(**DE44X_CONSTANTS)



from __future__ import annotations

import numpy as np
import quivr as qv

from ..time import Timestamp
from . import cartesian, cometary, keplerian
from .covariances import CoordinateCovariances, transform_covariances_jacobian
from .origin import Origin

__all__ = [
    "SphericalCoordinates",
]


class SphericalCoordinates(qv.Table):

    rho = qv.Float64Column(nullable=True)
    lon = qv.Float64Column(nullable=True)
    lat = qv.Float64Column(nullable=True)
    vrho = qv.Float64Column(nullable=True)
    vlon = qv.Float64Column(nullable=True)
    vlat = qv.Float64Column(nullable=True)
    time = Timestamp.as_column(nullable=True)
    covariance = CoordinateCovariances.as_column(nullable=True)
    origin = Origin.as_column()
    frame = qv.StringAttribute(default="unspecified")

    @property
    def values(self) -> np.ndarray:
        return np.array(
            self.table.select(["rho", "lon", "lat", "vrho", "vlon", "vlat"])
        )

    @property
    def sigma_rho(self):
        """
        1-sigma uncertainty in radial distance.
        """
        return self.covariance.sigmas[:, 0]

    @property
    def sigma_lon(self):
        """
        1-sigma uncertainty in longitude.
        """
        return self.covariance.sigmas[:, 1]

    @property
    def sigma_lat(self):
        """
        1-sigma uncertainty in latitude.
        """
        return self.covariance.sigmas[:, 2]

    @property
    def sigma_vrho(self):
        """
        1-sigma uncertainty in radial velocity.
        """
        return self.covariance.sigmas[:, 3]

    @property
    def sigma_vlon(self):
        """
        1-sigma uncertainty in longitudinal velocity.
        """
        return self.covariance.sigmas[:, 4]

    @property
    def sigma_vlat(self):
        """
        1-sigma uncertainty in latitudinal velocity.
        """
        return self.covariance.sigmas[:, 5]

    def to_unit_sphere(self, only_missing: bool = False) -> "SphericalCoordinates":
        """
        Convert to unit sphere. By default, all coordinates will have their rho values
        set to 1.0 and their vrho values set to 0.0. If only_missing is True, then only
        coordinates that have NaN values for rho will be set to 1.0 and coordinates that
        have NaN values for vrho will be set to 0.0.

        TODO: We could look at scaling the uncertainties as well, but this is not currently
        implemented nor probably necessary. This function will mostly be used to convert
        SphericalCoordinates that have missing radial distances to cartesian coordinates on a
        unit sphere.

        Parameters
        ----------
        only_missing : bool, optional
            If True, then only coordinates that have NaN values for rho will be set to 1.0 and
            coordinates that have NaN values for vrho will be set to 0.0. If False, then all
            coordinates will be set to 1.0 and 0.0, respectively. The default is False.

        Returns
        -------
        SphericalCoordinates
            Spherical coordinates on a unit sphere, with rho and vrho set to 1.0 and 0.0, respectively.
        """
        # Extract coordinate values
        coords = self.values

        # Set rho to 1.0 for all points that are NaN, or if force is True
        # then set rho to 1.0 for all points
        if not only_missing:
            mask = np.ones(len(coords), dtype=bool)
        else:
            mask = np.isnan(coords[:, 0])

        coords[mask, 0] = 1.0

        # Set vrho to 0.0 for all points that are NaN, or if force is True
        # then set vrho to 0.0 for all points
        if not only_missing:
            mask = np.ones(len(coords), dtype=bool)
        else:
            mask = np.isnan(coords[:, 3])

        coords[mask, 3] = 0.0

        # Convert back to spherical coordinates
        return SphericalCoordinates.from_kwargs(
            rho=coords[:, 0],
            lon=coords[:, 1],
            lat=coords[:, 2],
            vrho=coords[:, 3],
            vlon=coords[:, 4],
            vlat=coords[:, 5],
            time=self.time,
            covariance=self.covariance,
            origin=self.origin,
            frame=self.frame,
        )

    def to_cartesian(self) -> cartesian.CartesianCoordinates:
        from .transform import _spherical_to_cartesian, spherical_to_cartesian

        coords_cartesian = spherical_to_cartesian(self.values)
        coords_cartesian = np.array(coords_cartesian)

        if not self.covariance.is_all_nan():
            covariances_spherical = self.covariance.to_matrix()
            covariances_cartesian = transform_covariances_jacobian(
                self.values, covariances_spherical, _spherical_to_cartesian
            )
        else:
            covariances_cartesian = np.empty(
                (len(coords_cartesian), 6, 6), dtype=np.float64
            )
            covariances_cartesian.fill(np.nan)

        covariances_cartesian = CoordinateCovariances.from_matrix(covariances_cartesian)
        coords = cartesian.CartesianCoordinates.from_kwargs(
            x=coords_cartesian[:, 0],
            y=coords_cartesian[:, 1],
            z=coords_cartesian[:, 2],
            vx=coords_cartesian[:, 3],
            vy=coords_cartesian[:, 4],
            vz=coords_cartesian[:, 5],
            time=self.time,
            covariance=covariances_cartesian,
            origin=self.origin,
            frame=self.frame,
        )
        return coords

    @classmethod
    def from_cartesian(
        cls, cartesian: cartesian.CartesianCoordinates
    ) -> "SphericalCoordinates":
        from .transform import _cartesian_to_spherical, cartesian_to_spherical

        coords_spherical = cartesian_to_spherical(cartesian.values)
        coords_spherical = np.array(coords_spherical)

        if not cartesian.covariance.is_all_nan():
            cartesian_covariances = cartesian.covariance.to_matrix()
            covariances_spherical = transform_covariances_jacobian(
                cartesian.values, cartesian_covariances, _cartesian_to_spherical
            )
        else:
            covariances_spherical = np.empty(
                (len(coords_spherical), 6, 6), dtype=np.float64
            )
            covariances_spherical.fill(np.nan)

        covariances_spherical = CoordinateCovariances.from_matrix(covariances_spherical)
        coords = cls.from_kwargs(
            rho=coords_spherical[:, 0],
            lon=coords_spherical[:, 1],
            lat=coords_spherical[:, 2],
            vrho=coords_spherical[:, 3],
            vlon=coords_spherical[:, 4],
            vlat=coords_spherical[:, 5],
            time=cartesian.time,
            covariance=covariances_spherical,
            origin=cartesian.origin,
            frame=cartesian.frame,
        )

        return coords

    def to_cometary(self) -> cometary.CometaryCoordinates:
        return cometary.CometaryCoordinates.from_cartesian(self.to_cartesian())

    @classmethod
    def from_cometary(
        cls, cometary_coordinates: cometary.CometaryCoordinates
    ) -> SphericalCoordinates:
        return cls.from_cartesian(cometary_coordinates.to_cartesian())

    def to_keplerian(self) -> keplerian.KeplerianCoordinates:
        return keplerian.KeplerianCoordinates.from_cartesian(self.to_cartesian())

    @classmethod
    def from_keplerian(
        cls, keplerian_coordinates: keplerian.KeplerianCoordinates
    ) -> SphericalCoordinates:
        return cls.from_cartesian(keplerian_coordinates.to_cartesian())

    @classmethod
    def from_spherical(
        cls, spherical_coordinates: SphericalCoordinates
    ) -> SphericalCoordinates:
        return spherical_coordinates
